/*====================================================================
| Version: November 11, 2010
\===================================================================*/

/*====================================================================
| Ricard Delgado-Gonzalo
| EPFL/STI/IOA/LIB/BM.4.141
| e-mail: ricard.delgado@gmail.com
| URL: http://www.ricarddelgado.com/
\===================================================================*/
import ij.IJ;
import ij.ImagePlus;
import ij.ImageStack;
import ij.gui.GUI;
import ij.gui.Roi;
import ij.process.FloatProcessor;
import io.CTInput;
import io.CTOutput;

import java.awt.Choice;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;

import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JProgressBar;
import javax.swing.JScrollPane;
import javax.swing.JSpinner;
import javax.swing.JTextArea;
import javax.swing.SpinnerNumberModel;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.filechooser.FileNameExtensionFilter;

import display.CTLUT;
import display.CTOverlay;
import display.CTOverlay.CTCanvas;

import structures.CTConnectivity;
import structures.CTExtractedData;
import structures.CTRunParameters;
import structures.CTStatistics;

@SuppressWarnings("serial")
public class CellTracker_Displayer extends JDialog implements ActionListener, ItemListener, ChangeListener, Runnable {
	private static String appName = new String("Cell Tracker - Displayer");
	
	private static final int IDLEMODE = 0;
	private static final int RUNMODE  = 2;
	
	private String inputFolderCh1 = null;
	private String inputFolderCh2 = null;
	private String[] fileListCh1 = null;
	private String[] fileListCh2 = null;
	
	private Thread thread;
	
	//execution flags
	private boolean ch1Loaded;
	private boolean ch2Loaded;
	private int runmode;
	private boolean showOutline;	
	
	//execution parameters
	private CTRunParameters runParameters = null;
	private int Nstack;
	
	//interface
	private JTextArea taskLog = null;
	private JScrollPane taskPane = null;
	private JProgressBar progressBar = null;
	private JButton bnExit = null;
	private JButton bnContinue = null;
	private JButton bnCredits = null;
	private JButton bnLoadCh1 = null;
	private JButton bnLoadCh2 = null;
	private JButton bnSave = null;
	private JButton bnStats = null;
	private JButton bnLoad = null;
	private JLabel regionOffsetLabel = null;
	private JLabel sigmaNotMovingRegionLabel = null;
	private JLabel sigmaLabel = null;
	private JLabel upDistanceLabel = null;
	private JLabel downDistanceLabel = null;
	private JLabel lateralDistanceLabel = null;
	private JLabel lambdaLabel = null;
	private JLabel regularizationIterLabel = null;
	private SpinnerNumberModel regionOffsetModel = null;
	private SpinnerNumberModel sigmaNotMovingRegionModel = null;
	private SpinnerNumberModel sigmaModel = null;
	private SpinnerNumberModel upDisplacementModel = null;
	private SpinnerNumberModel downDisplacementModel = null;
	private SpinnerNumberModel lateralDisplacementModel = null;
	private SpinnerNumberModel lambdaModel = null;
	private SpinnerNumberModel regularizationIterModel = null;
	private JSpinner regionOffsetSpinner = null;
	private JSpinner sigmaNotMovingRegionSpinner = null; 
	private JSpinner sigmaSpinner = null; 
	private JSpinner upDisplacementSpinner = null; 
	private JSpinner downDisplacementSpinner = null; 
	private JSpinner lateralDisplacementSpinner = null; 
	private JSpinner lambdaSpinner = null; 
	private JSpinner regularizationIterSpinner = null;
	private Choice addremoveChooser;
	private Choice celltrackIDChooser;
	private Choice displayOptionsChooser;
	private JCheckBox showCellOutlineCheckBox = null;
	
	//extracted data
	private CTExtractedData[] extractedData = null;
	private CTConnectivity[] linkedData = null;
	
	//auxiliar classes
	CTOverlay overlayCh1;
	CTOverlay overlayCh2;
	private ImagePlus imCh1 = null;
	private ImagePlus imCh2 = null;
	private CTStatistics statisticsModule = null;
		
	public CellTracker_Displayer() {
		super(new JFrame(), appName);
		ch1Loaded = false;
		ch2Loaded = false;
		
		runmode = IDLEMODE;
		
		statisticsModule					= new CTStatistics();
		runParameters 						= new CTRunParameters();
		
		bnExit								= new JButton("Exit");
		bnContinue							= new JButton("Show");
		bnCredits	 						= new JButton("Credits");
		bnLoadCh1 							= new JButton("Select 1st channel");
		bnLoadCh2 							= new JButton("Select 2nd channel");
		bnLoad	 							= new JButton("Load results");
		bnStats								= new JButton("Show statistics");
		bnSave	 							= new JButton("Save results");
		
		regionOffsetLabel					= new JLabel("Region offset:");
		sigmaNotMovingRegionLabel			= new JLabel("Threshold NMR:");
		sigmaLabel							= new JLabel("Smoothing std:");
		lambdaLabel							= new JLabel("N-regularization:");
		upDistanceLabel 					= new JLabel("Max up displacement");
		downDistanceLabel 					= new JLabel("Max down displacement");
		lateralDistanceLabel				= new JLabel("Max lateral displacement");
		regularizationIterLabel				= new JLabel("N-regularization iter");
		
		showCellOutlineCheckBox				= new JCheckBox("Show cell outline");
		showOutline							= false;
		showCellOutlineCheckBox.setSelected(showOutline);
		
		runParameters.regionOffset			= 0;
		regionOffsetModel 					= new SpinnerNumberModel(runParameters.regionOffset, 0, Integer.MAX_VALUE, 1);
		
		runParameters.sigmaNotMovingRegion	= 0;
		sigmaNotMovingRegionModel 			= new SpinnerNumberModel(runParameters.sigmaNotMovingRegion, 0.0, 1.0, 0.001) ;
		
		runParameters.sigma 				= 0;
		sigmaModel 							= new SpinnerNumberModel(runParameters.sigma, 0.0, Double.POSITIVE_INFINITY, 0.001) ;
		
		runParameters.upDisplacement 		= 0;
		upDisplacementModel 				= new SpinnerNumberModel(runParameters.upDisplacement, 0.0, Double.POSITIVE_INFINITY, 1.0);

		runParameters.downDisplacement 		= 0;
		downDisplacementModel 				= new SpinnerNumberModel(runParameters.downDisplacement, 0.0, Double.POSITIVE_INFINITY, 1.0);
		
		runParameters.lateralDisplacement	= 0;
		lateralDisplacementModel 			= new SpinnerNumberModel(runParameters.lateralDisplacement, 0.0, Double.POSITIVE_INFINITY, 1.0);
		
		runParameters.lambda 				= 0;
		lambdaModel 						= new SpinnerNumberModel(runParameters.lambda, 0.0, Double.POSITIVE_INFINITY, 0.001);
		
		runParameters.regularizationIter	= 0;
		regularizationIterModel				= new SpinnerNumberModel(runParameters.regularizationIter, 0, Integer.MAX_VALUE, 1);
		
		regionOffsetSpinner					= new JSpinner(regionOffsetModel);
		sigmaNotMovingRegionSpinner			= new JSpinner(sigmaNotMovingRegionModel);
		sigmaSpinner 						= new JSpinner(sigmaModel);
		upDisplacementSpinner 				= new JSpinner(upDisplacementModel);
		downDisplacementSpinner 			= new JSpinner(downDisplacementModel);
		lateralDisplacementSpinner			= new JSpinner(lateralDisplacementModel);
		lambdaSpinner 						= new JSpinner(lambdaModel);
		regularizationIterSpinner			= new JSpinner(regularizationIterModel);
		
		regionOffsetSpinner.setEnabled(false);
		sigmaNotMovingRegionSpinner.setEnabled(false);
		sigmaSpinner.setEnabled(false); 						
		upDisplacementSpinner.setEnabled(false); 				
		downDisplacementSpinner.setEnabled(false); 			
		lateralDisplacementSpinner.setEnabled(false);			
		lambdaSpinner.setEnabled(false); 						
		regularizationIterSpinner.setEnabled(false);			
		
		addremoveChooser 					= new Choice();
		addremoveChooser.add("Disable edit mode");
		addremoveChooser.add("Remove connection");
		addremoveChooser.add("Add connection");
		
		celltrackIDChooser					= new Choice();
		celltrackIDChooser.add("Do not show ID");
		celltrackIDChooser.add("Show cell ID");
		celltrackIDChooser.add("Show track ID");
		
		displayOptionsChooser				= new Choice();
		displayOptionsChooser.add("Do not display tracking");
		displayOptionsChooser.add("Overlay pairing");
		displayOptionsChooser.add("Overlay tracks");
		
		
		regionOffsetSpinner.setPreferredSize(new Dimension(110,22));
		sigmaNotMovingRegionSpinner.setPreferredSize(new Dimension(110,22));
		sigmaSpinner.setPreferredSize(new Dimension(110,22));
		upDisplacementSpinner.setPreferredSize(new Dimension(110,22));
		downDisplacementSpinner.setPreferredSize(new Dimension(110,22));
		lateralDisplacementSpinner.setPreferredSize(new Dimension(110,22));
		lambdaSpinner.setPreferredSize(new Dimension(110,22));
		regularizationIterSpinner.setPreferredSize(new Dimension(110,22));
		
		taskLog 	= new JTextArea();
        taskLog.setEditable(false);
        taskPane 	= new JScrollPane(taskLog);
		taskPane.setPreferredSize(new Dimension(600, 140));
        
		progressBar	= new JProgressBar();
	    progressBar.setValue(0);
	    progressBar.setStringPainted(true);
	    progressBar.setVisible(false);
	   
		addComponentsToInterface();
		
		showCellOutlineCheckBox.setEnabled(false);
		addremoveChooser.setEnabled(false);
		celltrackIDChooser.setEnabled(false);
		displayOptionsChooser.setEnabled(false);
		bnContinue.setEnabled(false);
		bnSave.setEnabled(false);
		bnStats.setEnabled(false);
		
		bnStats.addActionListener(this);
		bnContinue.addActionListener(this);
		bnExit.addActionListener(this);
		bnCredits.addActionListener(this);
		bnLoadCh1.addActionListener(this);
		bnLoadCh2.addActionListener(this);
		bnLoad.addActionListener(this);
		bnSave.addActionListener(this);
		showCellOutlineCheckBox.addItemListener(this);
		addremoveChooser.addItemListener(this);
		celltrackIDChooser.addItemListener(this);
		displayOptionsChooser.addItemListener(this);
		regionOffsetSpinner.addChangeListener(this);
		sigmaNotMovingRegionSpinner.addChangeListener(this);
		sigmaSpinner.addChangeListener(this);
		upDisplacementSpinner.addChangeListener(this);
		downDisplacementSpinner.addChangeListener(this);
		lateralDisplacementSpinner.addChangeListener(this);
		lambdaSpinner.addChangeListener(this);
		regularizationIterSpinner.addChangeListener(this);
		
		writeLog("Please, select the source folder");
		
		GUI.center(this);
		setPreferredSize(new Dimension(750,450));
		setResizable(false);
		pack();
		GUI.center(this);
		setVisible(true);
	}
	

	public synchronized void actionPerformed(ActionEvent e) {
		if (e.getSource() == bnExit) {
			thread = null;
			if(imCh1!=null) imCh1.close();
			if(imCh2!=null) imCh2.close();
			statisticsModule.disposeAll();
			dispose();
		}else if(e.getSource() == bnStats){
			statisticsModule.segmentationPerformance(extractedData);
		}else if(e.getSource() == bnCredits){
			new CTCredits(null);
		}else if(runmode == IDLEMODE){
			if (e.getSource() == bnContinue) {
				if (thread == null) {
					runmode = RUNMODE;
					bnContinue.setEnabled(false);
					thread = new Thread(this);
					thread.setPriority(Thread.MIN_PRIORITY);
					thread.start();
				}
			}else if(e.getSource() == bnSave){
				saveResults();
			}else if(e.getSource() == bnLoadCh1) {
				if(readFrames(0)){
					ch1Loaded = true;
					if(ch2Loaded && ch1Loaded){
						bnLoadCh1.setEnabled(false);
						bnLoadCh2.setEnabled(false);
						if (fileListCh1.length==fileListCh2.length){
							Nstack = fileListCh1.length;
							bnContinue.setEnabled(true);
							writeLog("Stacks selected, please select a results file and press the Show button");
						}else{
							writeLog("Stack sizes do not match");
							writeLog("Please, select stacks with the same size");
							bnLoadCh1.setEnabled(true);
							bnLoadCh2.setEnabled(true);
						}
					}
				}
			}else if(e.getSource() == bnLoadCh2) {
				if(readFrames(1)){
					ch2Loaded = true;
					if(ch2Loaded && ch1Loaded){
						bnLoadCh1.setEnabled(false);
						bnLoadCh2.setEnabled(false);
						if (fileListCh1.length==fileListCh2.length){
							Nstack = fileListCh1.length;
							bnContinue.setEnabled(true);
							writeLog("Stacks selected, please select a results file and press the Show button");
						}else{
							writeLog("Stack sizes do not match");
							writeLog("Please, select stacks with the same size");
							bnLoadCh1.setEnabled(true);
							bnLoadCh2.setEnabled(true);
						}
					}			
				}
			}else if(e.getSource() == bnLoad){
				if(ch1Loaded && ch2Loaded){
					if(loadResults()){
						regionOffsetSpinner.setValue(runParameters.regionOffset);
						sigmaNotMovingRegionSpinner.setValue(runParameters.sigmaNotMovingRegion);			
						sigmaSpinner.setValue(runParameters.sigma);
						upDisplacementSpinner.setValue(runParameters.upDisplacement);		
						downDisplacementSpinner.setValue(runParameters.downDisplacement);
						lateralDisplacementSpinner.setValue(runParameters.lateralDisplacement);
						lambdaSpinner.setValue(runParameters.lambda);
						regularizationIterSpinner.setValue(runParameters.regularizationIter);
						bnLoad.setEnabled(false);
					}
				}else{
					writeLog("Image stacks not selected, please select the image stacks before loading the results file");
				}
			}
		}
	}
	
	private boolean loadResults(){
		JFrame frame = new JFrame();
		JFileChooser fc = new JFileChooser();
		FileNameExtensionFilter filter = new FileNameExtensionFilter("XML Files", "xml");
		fc.setFileFilter(filter);
		int returnVal = fc.showOpenDialog(frame);
		if(returnVal == JFileChooser.APPROVE_OPTION) {
			writeLog("Parsing "+fc.getSelectedFile().getAbsolutePath()+"...");
			extractedData = new CTExtractedData[Nstack];
			linkedData = new CTConnectivity[Nstack];
			bnLoad.setEnabled(false);
			boolean goodLoad = CTInput.loadXML(fc.getSelectedFile().getAbsolutePath(), runParameters, extractedData, linkedData);
			if(!goodLoad){
				writeLog("Selected file corrupted");
			}else{
				writeLog("File "+fc.getSelectedFile().getAbsolutePath()+" parsed successfully");
					
			}
			return goodLoad;
		}else{
			return false;
		}
	}
	
	
	private void saveResults() {
		JFrame frame = new JFrame();
		JFileChooser fc = new JFileChooser();
		FileNameExtensionFilter filter = new FileNameExtensionFilter("XML Files", "xml");
		fc.setFileFilter(filter);
		int returnVal = fc.showSaveDialog(frame);
		if(returnVal == JFileChooser.APPROVE_OPTION) {
			CTOutput.saveXML(fc.getSelectedFile().getAbsolutePath(), runParameters, extractedData, linkedData);
			CTOutput.saveSingleCell(fc.getSelectedFile().getAbsolutePath(), extractedData);
			CTOutput.saveSingleCellOutline(fc.getSelectedFile().getAbsolutePath(), extractedData);
		}
	}

	private boolean readFrames(int channel){
		JFrame frame = new JFrame();
		JFileChooser fc = new JFileChooser();
		fc.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
		
		if(channel==0){
			int returnVal = fc.showOpenDialog(frame);
			if(returnVal == JFileChooser.APPROVE_OPTION) {
				inputFolderCh1 = fc.getSelectedFile().getPath();
				fileListCh1 = CTInput.getListDir(inputFolderCh1);
			}
			if(fileListCh1.length>0){
				writeLog(fileListCh1.length +" files detected in folder "+inputFolderCh1);
				return true;
			}else{
				writeLog("The selected folder did not contain any valid image");
				writeLog("Please select another folder");
				return false;
			}
		}else if(channel==1){
			int returnVal = fc.showOpenDialog(frame);
			if(returnVal == JFileChooser.APPROVE_OPTION) {
				inputFolderCh2 = fc.getSelectedFile().getPath();
				fileListCh2 = CTInput.getListDir(inputFolderCh2);
			}
			if(fileListCh2.length>0){
				writeLog(fileListCh2.length +" files detected in folder "+inputFolderCh2);
				return true;
			}else{
				writeLog("The selected folder did not contain any valid image");
				writeLog("Please select another folder");
				return false;
			}
		}else{
			IJ.error("readFrames internal error");
			return false;
		}
	}
	
	private void addComponentsToInterface(){
		GridBagLayout gridbag = new GridBagLayout();
		GridBagConstraints c = new GridBagConstraints();
		setLayout(gridbag);
		
		JPanel displayPanel = new JPanel(new GridBagLayout());
		displayPanel.setBorder(BorderFactory.createTitledBorder("Display"));
		displayPanel.setPreferredSize(new Dimension(235, 240));
		
		c.insets = new Insets(0,0,0,0);
		c.gridwidth = 1;
		c.gridheight = 1;
		c.fill = GridBagConstraints.HORIZONTAL;
		c.gridx = 0;
		c.gridy = 0;
		gridbag.setConstraints(showCellOutlineCheckBox, c);
		displayPanel.add(showCellOutlineCheckBox, c);		
		
		c.insets = new Insets(0,0,0,0);
		c.gridwidth = 1;
		c.gridheight = 1;
		c.fill = GridBagConstraints.HORIZONTAL;
		c.gridx = 0;
		c.gridy = 1;
		gridbag.setConstraints(celltrackIDChooser, c);
		displayPanel.add(celltrackIDChooser, c);		
		
		c.insets = new Insets(0,0,0,0);
		c.gridwidth = 1;
		c.gridheight = 1;
		c.fill = GridBagConstraints.HORIZONTAL;
		c.gridx = 0;
		c.gridy = 2;
		gridbag.setConstraints(displayOptionsChooser, c);
		displayPanel.add(displayOptionsChooser, c);		
		
		JPanel actionsPanel = new JPanel(new GridBagLayout());
		actionsPanel.setBorder(BorderFactory.createTitledBorder("Actions"));
		actionsPanel.setPreferredSize(new Dimension(205, 240));
		
		c.insets = new Insets(0,0,0,0);
		c.gridwidth = 1;
		c.gridheight = 1;
		c.fill = GridBagConstraints.HORIZONTAL;
		c.gridx = 0;
		c.gridy = 0;
		gridbag.setConstraints(bnLoadCh1, c);
		actionsPanel.add(bnLoadCh1, c);
		
		c.insets = new Insets(0,0,0,0);
		c.gridwidth = 1;
		c.gridheight = 1;
		c.fill = GridBagConstraints.HORIZONTAL;
		c.gridx = 0;
		c.gridy = 1;
		gridbag.setConstraints(bnLoadCh2, c);
		actionsPanel.add(bnLoadCh2, c);

		c.insets = new Insets(0,0,0,0);
		c.gridwidth = 1;
		c.gridheight = 1;
		c.fill = GridBagConstraints.HORIZONTAL;
		c.gridx = 0;
		c.gridy = 2;
		gridbag.setConstraints(bnLoad, c);
		actionsPanel.add(bnLoad, c);
		
		c.insets = new Insets(0,0,0,0);
		c.gridwidth = 1;
		c.gridheight = 1;
		c.fill = GridBagConstraints.HORIZONTAL;
		c.gridx = 0;
		c.gridy = 3;
		gridbag.setConstraints(bnStats, c);
		actionsPanel.add(bnStats, c);
		
		c.insets = new Insets(0,0,0,0);
		c.gridwidth = 1;
		c.gridheight = 1;
		c.fill = GridBagConstraints.HORIZONTAL;
		c.gridx = 0;
		c.gridy = 4;
		gridbag.setConstraints(bnSave, c);
		actionsPanel.add(bnSave, c);
		
		c.insets = new Insets(0,0,0,0);
		c.gridwidth = 1;
		c.gridheight = 1;
		c.fill = GridBagConstraints.HORIZONTAL;
		c.gridx = 0;
		c.gridy = 5;
		gridbag.setConstraints(bnCredits, c);
		actionsPanel.add(bnCredits, c);
		
		c.insets = new Insets(0,0,0,0);
		c.gridwidth = 1;
		c.gridheight = 1;
		c.fill = GridBagConstraints.HORIZONTAL;
		c.gridx = 0;
		c.gridy = 6;
		gridbag.setConstraints(addremoveChooser, c);
		actionsPanel.add(addremoveChooser, c);
		
		JPanel settingsPanel = new JPanel(new GridBagLayout());
		settingsPanel.setBorder(BorderFactory.createTitledBorder("Settings"));
		settingsPanel.setPreferredSize(new Dimension(285, 240));
		
		c.insets = new Insets(0,0,0,0);
		c.gridwidth = 1;
		c.gridheight = 1;
		c.anchor = GridBagConstraints.WEST;
		c.gridx = 0;
		c.gridy = 0;
		gridbag.setConstraints(regionOffsetLabel, c);
		settingsPanel.add(regionOffsetLabel, c);
		
		c.insets = new Insets(0,0,0,0);
		c.gridwidth = 1;
		c.gridheight = 1;
		c.anchor = GridBagConstraints.WEST;
		c.gridx = 1;
		c.gridy = 0;
		gridbag.setConstraints(regionOffsetSpinner, c);
		settingsPanel.add(regionOffsetSpinner, c);
		
		c.insets = new Insets(0,0,0,0);
		c.gridwidth = 1;
		c.gridheight = 1;
		c.anchor = GridBagConstraints.WEST;
		c.gridx = 0;
		c.gridy = 1;
		gridbag.setConstraints(sigmaNotMovingRegionLabel, c);
		settingsPanel.add(sigmaNotMovingRegionLabel, c);

		c.insets = new Insets(0,0,0,0);
		c.gridwidth = 1;
		c.gridheight = 1;
		c.anchor = GridBagConstraints.WEST;
		c.gridx = 1;
		c.gridy = 1;
		gridbag.setConstraints(sigmaNotMovingRegionSpinner, c);
		settingsPanel.add(sigmaNotMovingRegionSpinner, c);

		c.insets = new Insets(0,0,0,0);
		c.gridwidth = 1;
		c.gridheight = 1;
		c.anchor = GridBagConstraints.WEST;
		c.gridx = 0;
		c.gridy = 2;
		gridbag.setConstraints(sigmaLabel, c);
		settingsPanel.add(sigmaLabel, c);

		c.insets = new Insets(0,0,0,0);
		c.gridwidth = 1;
		c.gridheight = 1;
		c.anchor = GridBagConstraints.WEST;
		c.gridx = 1;
		c.gridy = 2;
		gridbag.setConstraints(sigmaSpinner, c);
		settingsPanel.add(sigmaSpinner, c);

		c.insets = new Insets(0,0,0,0);
		c.gridwidth = 1;
		c.gridheight = 1;
		c.anchor = GridBagConstraints.WEST;
		c.gridx = 0;
		c.gridy = 3;
		gridbag.setConstraints(lambdaLabel, c);
		settingsPanel.add(lambdaLabel, c);

		c.insets = new Insets(0,0,0,0);
		c.gridwidth = 1;
		c.gridheight = 1;
		c.anchor = GridBagConstraints.WEST;
		c.gridx = 1;
		c.gridy = 3;
		gridbag.setConstraints(lambdaSpinner, c);
		settingsPanel.add(lambdaSpinner, c);
		
		c.insets = new Insets(0,0,0,0);
		c.gridwidth = 1;
		c.gridheight = 1;
		c.anchor = GridBagConstraints.WEST;
		c.gridx = 0;
		c.gridy = 4;
		gridbag.setConstraints(regularizationIterLabel, c);
		settingsPanel.add(regularizationIterLabel, c);

		c.insets = new Insets(0,0,0,0);
		c.gridwidth = 1;
		c.gridheight = 1;
		c.anchor = GridBagConstraints.WEST;
		c.gridx = 1;
		c.gridy = 4;
		gridbag.setConstraints(regularizationIterSpinner, c);
		settingsPanel.add(regularizationIterSpinner, c);
		
		c.insets = new Insets(0,0,0,0);
		c.anchor=GridBagConstraints.WEST;
		c.gridx = 0;
		c.gridy = 5;
		gridbag.setConstraints(upDistanceLabel, c);
		settingsPanel.add(upDistanceLabel, c);
		
		c.insets = new Insets(0,0,0,0);
		c.anchor=GridBagConstraints.WEST;
		c.gridx = 1;
		c.gridy = 5;
		gridbag.setConstraints(upDisplacementSpinner, c);
		settingsPanel.add(upDisplacementSpinner, c);
		
		c.insets = new Insets(0,0,0,0);
		c.anchor=GridBagConstraints.WEST;
		c.gridx = 0;
		c.gridy = 6;
		gridbag.setConstraints(downDistanceLabel, c);
		settingsPanel.add(downDistanceLabel, c);
		
		c.insets = new Insets(0,0,0,0);
		c.anchor=GridBagConstraints.WEST;
		c.gridx = 1;
		c.gridy = 6;
		gridbag.setConstraints(downDisplacementSpinner, c);
		settingsPanel.add(downDisplacementSpinner, c);
		
		
		c.insets = new Insets(0,0,0,0);
		c.anchor=GridBagConstraints.WEST;
		c.gridx = 0;
		c.gridy = 7;
		gridbag.setConstraints(lateralDistanceLabel, c);
		settingsPanel.add(lateralDistanceLabel, c);
		
		c.insets = new Insets(0,0,0,0);
		c.anchor=GridBagConstraints.WEST;
		c.gridx = 1;
		c.gridy = 7;
		gridbag.setConstraints(lateralDisplacementSpinner, c);
		settingsPanel.add(lateralDisplacementSpinner, c);
		
		//Build general panel
		c.insets = new Insets(0,0,0,0);
		c.gridwidth = 1;
		c.gridheight = 1;
		c.gridx = 0;
		c.gridy = 0;
		gridbag.setConstraints(settingsPanel, c);
		add(settingsPanel);
		
		c.insets = new Insets(0,0,0,0);
		c.gridx = 1;
		c.gridy = 0;
		gridbag.setConstraints(displayPanel, c);
		add(displayPanel);
		
		c.insets = new Insets(0,0,0,0);
		c.gridx = 2;
		c.gridy = 0;
		gridbag.setConstraints(actionsPanel, c);
		add(actionsPanel);
		
		c.insets = new Insets(2,2,2,3);
		c.anchor=GridBagConstraints.CENTER;
		c.gridwidth = 3;
		c.gridheight = 1;
		c.gridx = 0;
		c.gridy = 1;
		gridbag.setConstraints(taskPane, c);
		add(taskPane);
		
		c.insets = new Insets(0,0,0,0);
		c.anchor=GridBagConstraints.WEST;
		c.gridwidth = 1;
		c.gridheight = 1;
		c.gridx = 0;
		c.gridy = 2;
		gridbag.setConstraints(bnContinue, c);
		add(bnContinue);
		
		c.insets = new Insets(0,0,0,0);
		c.anchor=GridBagConstraints.WEST;
		c.gridwidth = 2;
		c.gridheight = 1;
		c.gridx = 0;
		c.gridy = 2;
		gridbag.setConstraints(progressBar, c);
		add(progressBar);
		
		
		c.insets = new Insets(0,0,0,0);
		c.anchor = GridBagConstraints.EAST;
		c.gridwidth = 1;
		c.gridheight = 1;
		c.gridx = 2;
		c.gridy = 2;
		gridbag.setConstraints(bnExit, c);
		add(bnExit, c);
	}
	
	private void writeLog(String str){
		if(taskLog!=null){
			taskLog.append(str+"\n");
			taskLog.setCaretPosition(taskLog.getDocument().getLength());
		}else{
			IJ.log(str);
		}
	}
	
	private void loadStack(Roi roi){
		progressBar.setMinimum(0);
		progressBar.setMaximum(Nstack-1);
		progressBar.setVisible(true);
		bnContinue.setVisible(false);
		
		FloatProcessor fpch1 = CTInput.readProcessor(inputFolderCh1, fileListCh1[0], roi);
		FloatProcessor fpch2 = CTInput.readProcessor(inputFolderCh2, fileListCh2[0], roi);

		CTLUT.setLUT(fpch2, CTLUT.GREEN);
		
		int nx, ny;
		if(roi==null){
			nx = fpch1.getWidth();
			ny = fpch1.getHeight();
		}else{
			nx = roi.getBounds().width;
			ny = roi.getBounds().height;
		}
		
		ImageStack originalStackCh1 = new ImageStack(nx, ny);
		ImageStack originalStackCh2 = new ImageStack(nx, ny);
		originalStackCh1.addSlice("", fpch1);
		originalStackCh2.addSlice("", fpch2);
	
		for(int frame=1; frame<Nstack; frame++){
			progressBar.setValue(frame-1);
			fpch1 = CTInput.readProcessor(inputFolderCh1, fileListCh1[frame], roi);
			fpch2 = CTInput.readProcessor(inputFolderCh2, fileListCh2[frame], roi);
			CTLUT.setLUT(fpch2, CTLUT.GREEN);
			originalStackCh1.addSlice("", fpch1);
			originalStackCh2.addSlice("", fpch2);
		}
		
		bnContinue.setVisible(true);
		progressBar.setVisible(false);
		writeLog("Stack of "+ Nstack + " files loaded");
	
		imCh1 = new ImagePlus("Segmentation Window - Channel 1", originalStackCh1);
		imCh2 = new ImagePlus("Segmentation Window - Channel 2", originalStackCh2);
		imCh1.show();
		imCh2.show();
	}
	
	public void run() {
		Cursor cursor = getCursor();
		setCursor(new Cursor(Cursor.WAIT_CURSOR));
		
		switch (runmode) {
			case RUNMODE:
				bnContinue.setVisible(false);
				progressBar.setMinimum(0);
				progressBar.setMaximum(Nstack-1);
				progressBar.setVisible(true);
				displayResults(new Roi(runParameters.roiBox));
				
				writeLog("Execution finished");
				showCellOutlineCheckBox.setEnabled(true);
				addremoveChooser.setEnabled(true);
				celltrackIDChooser.setEnabled(true);
				displayOptionsChooser.setEnabled(true);
				bnContinue.setEnabled(false);
				bnContinue.setVisible(true);
				progressBar.setVisible(false);
				bnSave.setEnabled(true);
				bnStats.setEnabled(true);
				break;
			default:
				break;
		}
		runmode = IDLEMODE;
		setCursor(cursor);
		thread = null;
	}
	
	private void displayResults(Roi roi){
		loadStack(roi);
		overlayCh1 = new CTOverlay(imCh1, extractedData, linkedData);
		overlayCh2 = new CTOverlay(imCh2, extractedData, linkedData);
		overlayCh1.setOtherChannel(overlayCh2);
		overlayCh2.setOtherChannel(overlayCh1);
		overlayCh1.canvas.setEditMode(CTCanvas.IDLESTATE);
		overlayCh2.canvas.setEditMode(CTCanvas.IDLESTATE);
		overlayCh1.canvas.showCellOutline(showOutline);
		overlayCh2.canvas.showCellOutline(showOutline);
	}
	

	public void stateChanged(ChangeEvent e) {
		if(e.getSource() == sigmaSpinner){
			runParameters.sigma = ((Double)sigmaSpinner.getValue()).doubleValue();
		}else if(e.getSource() == upDisplacementSpinner){
			runParameters.upDisplacement = ((Double)upDisplacementSpinner.getValue()).doubleValue();
		}else if(e.getSource() == downDisplacementSpinner){
			runParameters.downDisplacement = ((Double)downDisplacementSpinner.getValue()).doubleValue();
		}else if(e.getSource() == lateralDisplacementSpinner){
			runParameters.lateralDisplacement = ((Double)lateralDisplacementSpinner.getValue()).doubleValue();
		}else if(e.getSource() == lambdaSpinner){
			runParameters.regionOffset = ((Integer)regionOffsetSpinner.getValue()).intValue();
		}else if(e.getSource() == sigmaNotMovingRegionSpinner){
			runParameters.sigmaNotMovingRegion = ((Double)sigmaNotMovingRegionSpinner.getValue()).doubleValue();
		}else if(e.getSource() == regularizationIterSpinner){
			runParameters.regularizationIter = ((Integer)regularizationIterSpinner.getValue()).intValue();
		}
	}
	
	public void itemStateChanged(ItemEvent e) {
		if(e.getSource()==addremoveChooser) {
			switch (addremoveChooser.getSelectedIndex()) {
			case 0:
				writeLog("Edit mode disabled");
				overlayCh1.canvas.setEditMode(CTCanvas.IDLESTATE);
				overlayCh2.canvas.setEditMode(CTCanvas.IDLESTATE);
				break;
			case 1:
				writeLog("Edit mode enabled");
				writeLog("Remove connection mode set");
				overlayCh1.canvas.setEditMode(CTCanvas.REMOVECONECTIONSTATE);
				overlayCh2.canvas.setEditMode(CTCanvas.REMOVECONECTIONSTATE);
				break;
			case 2:
				writeLog("Edit mode enabled");
				writeLog("Add connection mode set");
				overlayCh1.canvas.setEditMode(CTCanvas.ADDCONECTIONSTATE);
				overlayCh2.canvas.setEditMode(CTCanvas.ADDCONECTIONSTATE);
				break;
			default:
				break;
			}
		}else if(e.getSource()==celltrackIDChooser){
			switch (celltrackIDChooser.getSelectedIndex()) {
			case 0:
				writeLog("Hiding ID's");
				overlayCh1.canvas.setLabelMode(CTCanvas.NOLABELMODE);
				overlayCh2.canvas.setLabelMode(CTCanvas.NOLABELMODE);
				break;
			case 1:
				writeLog("Displaying cell ID's");
				overlayCh1.canvas.setLabelMode(CTCanvas.CELLLABELMODE);
				overlayCh2.canvas.setLabelMode(CTCanvas.CELLLABELMODE);
				break;
			case 2:
				writeLog("Displaying track ID's");
				overlayCh1.canvas.setLabelMode(CTCanvas.TRACKLABELMODE);
				overlayCh2.canvas.setLabelMode(CTCanvas.TRACKLABELMODE);
				break;
			default:
				break;
			}
		}else if(e.getSource()==displayOptionsChooser){
			switch (displayOptionsChooser.getSelectedIndex()) {
			case 0:
				writeLog("Hiding tracking information");
				overlayCh1.canvas.setDisplayMode(CTCanvas.NOTRACKINGMODE);
				overlayCh2.canvas.setDisplayMode(CTCanvas.NOTRACKINGMODE);
				break;
			case 1:
				writeLog("Displaying pairing infomation");
				overlayCh1.canvas.setDisplayMode(CTCanvas.PAIRINGMODE);
				overlayCh2.canvas.setDisplayMode(CTCanvas.PAIRINGMODE);
				break;
			case 2:
				writeLog("Displaying track information");
				overlayCh1.canvas.setDisplayMode(CTCanvas.TRACKMODE);
				overlayCh2.canvas.setDisplayMode(CTCanvas.TRACKMODE);
				break;
			default:
				break;
			}
		}
					
		Object source = e.getItemSelectable();
	    if(source == showCellOutlineCheckBox){
	    	showOutline =! showOutline;
	    	overlayCh1.canvas.showCellOutline(showOutline);
			overlayCh2.canvas.showCellOutline(showOutline);
        }
	}
} 

