/*====================================================================
| Version: November 11, 2010
\===================================================================*/

/*====================================================================
| Ricard Delgado-Gonzalo
| EPFL/STI/IOA/LIB/BM.4.141
| e-mail: ricard.delgado@gmail.com
| URL: http://www.ricarddelgado.com/
\===================================================================*/
package display;

import ij.ImagePlus;
import ij.gui.ImageCanvas;
import ij.gui.StackWindow;
import ij.gui.Toolbar;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Rectangle;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.Polygon;

import structures.CTConnectivity;
import structures.CTExtractedData;

public class CTOverlay {
	
	public CTCanvas canvas;
	StackWindow stackWindow = null;
	private CTExtractedData[] extractedData;
	private CTConnectivity[] linkedData;
	private int Nstack;
	private int labelTmp;
	private CTOverlay otherChannel = null;
	
	
	public CTOverlay(ImagePlus imageplus, CTExtractedData[] extractedData, CTConnectivity[] linkedData) {
		this.extractedData = extractedData;
		this.linkedData = linkedData;
		Nstack = extractedData.length;
		canvas = new CTCanvas(imageplus);
		stackWindow = new StackWindow(imageplus, canvas);
	}
	
	public void setOtherChannel(CTOverlay otherChannel){
		this.otherChannel = otherChannel;
	}
	
	private void reLabelTracks(){
		int Nregions;
		//initialize
		for(int i=0; i<Nstack; i++){
			Nregions = linkedData[i].Nregions;
			linkedData[i].trackLabel = new int[Nregions];
			for(int j=0; j<Nregions; j++){
				linkedData[i].trackLabel[j] = -1;
			}
		}
		
		//label in recursive way
		labelTmp = 0;
		for(int frame=0; frame<Nstack; frame++){
			Nregions = linkedData[frame].Nregions;
			for(int cell=0; cell<Nregions; cell++){
				if(linkedData[frame].trackLabel[cell]==-1){ 
					labelTrack(frame, cell, labelTmp);
					labelTmp++;
				}
			}
		}
	}	
	
	private void labelTrack(int frame, int cell, int label) {
		linkedData[frame].trackLabel[cell] = label;
		int Ndaughters = linkedData[frame].children[cell].size();
		if(Ndaughters != 0){
			for(int i=0; i<Ndaughters; i++){
				labelTrack(frame+1, linkedData[frame].children[cell].elementAt(i), label);
			}
		}		
	}
	
	public void dispose(){
        stackWindow.dispose();
	}
	
    @SuppressWarnings("serial")
	public class CTCanvas extends ImageCanvas implements MouseListener{
    	public static final int IDLESTATE = 0;
    	public static final int REMOVECONECTIONSTATE = 1;
    	public static final int ADDCONECTIONSTATE = 2;
    	
    	public static final int NOLABELMODE = 0;
    	public static final int TRACKLABELMODE = 1;
    	public static final int CELLLABELMODE = 2;
    	
    	public static final int NOTRACKINGMODE = 0;
    	public static final int PAIRINGMODE = 1;
    	public static final int TRACKMODE = 2;
    	
    	private static final int MAXCOLORS = 500;
    	
    	private Color[] trackLUT = null;
    	
    	private double mag;
		private Rectangle rectOffset; 
    	private int currentFrame;
    	private Color centerColor = Color.yellow;
    	private Color[] outlineColor = null;
    	private Color pairingColor = Color.red;
    	private int editMode = IDLESTATE;
    	private int labelMode = NOLABELMODE;
    	private int displayMode = NOTRACKINGMODE;
    	private int cellNumberOld = -1;
    	private int frameOld;
    	private boolean showOutline = false;
    	
    	
		public CTCanvas(ImagePlus imp) {
            super(imp);
            this.imp = imp;
            //look for the max number of cells in a given frame
            int nMaxCells = 0;
            for(int i=0; i<Nstack; i++){
            	int nRegions = extractedData[i].getNregions();
            	if (nRegions>nMaxCells){
            		nMaxCells = nRegions;
            	}
            }
            trackLUT = CTLUT.createLUTColor(nMaxCells*Nstack);
            outlineColor = CTLUT.createLUTColor(nMaxCells);
        }

		private void updateWindows(){
			repaint();
			otherChannel.canvas.repaint();
		}
		
		public void setLabelMode(int mode){
			labelMode = mode;
			updateWindows();
		}
		
		public void setDisplayMode(int mode){
			displayMode = mode;
			updateWindows();
		}
		
		
		public void showCellOutline(boolean showOutline){
			this.showOutline = showOutline;
			updateWindows();
		}
		
		public void mouseClicked (MouseEvent e)	{
			super.mousePressed(e);
			if (Toolbar.getToolId() == Toolbar.CROSSHAIR){		
				int x = canvas.offScreenX(e.getX());
				int y = canvas.offScreenY(e.getY());
				currentFrame = imp.getSlice()-1;
				
				 switch(editMode){
		            case ADDCONECTIONSTATE:
						if ((e.getModifiers() & MouseEvent.ALT_MASK) == MouseEvent.ALT_MASK){
							if(currentFrame == frameOld){
								int cellNumber = Math.round(Float.intBitsToFloat(extractedData[currentFrame+1].getRegionLabelMap().get(x, y)))-11;
								if(cellNumber>=0){
									if(linkedData[currentFrame+1].mother[cellNumber]==-1){
										linkedData[currentFrame+1].mother[cellNumber] = cellNumberOld;
										linkedData[currentFrame].children[cellNumberOld].addElement(new Integer(cellNumber));
										reLabelTracks();
									}
								}
							}
						}else{
							int cellNumber = Math.round(Float.intBitsToFloat(extractedData[currentFrame].getRegionLabelMap().get(x, y)))-11;
							if(cellNumber>=0){
								if(linkedData[currentFrame].children[cellNumber].size()==0){
									cellNumberOld = cellNumber;
									frameOld = currentFrame;
								}else{
									cellNumberOld = -1;
									frameOld = -1;
								}
							}else{
								cellNumberOld = -1;
								frameOld = -1;
							}
						}
						break;
		            case REMOVECONECTIONSTATE:
						//remove connection
						cellNumberOld = -1;
						frameOld = -1;
						int cellNumber = Math.round(Float.intBitsToFloat(extractedData[currentFrame].getRegionLabelMap().get(x, y)))-11;
						if(cellNumber>=0){
							if(linkedData[currentFrame].children[cellNumber].size()>0){
								int Ndaughters = linkedData[currentFrame].children[cellNumber].size();
								for(int i=Ndaughters-1; i>=0; i--){
									int daug = linkedData[currentFrame].children[cellNumber].elementAt(i);
									linkedData[currentFrame].children[cellNumber].removeElementAt(i);
									linkedData[currentFrame+1].mother[daug]=-1;
								}
								reLabelTracks();
							}
						}					
		            	break;
		            default:
		            	int cellId = Math.round(Float.intBitsToFloat(extractedData[currentFrame+1].getRegionLabelMap().get(x, y)))-11;
						if(cellId>=0){
							extractedData[currentFrame].printCellInfo(cellId);
						}
		            	break;
				 }
			}
		}
		public void mouseEntered (MouseEvent e)	{}
		public void mouseExited  (MouseEvent e)	{}
		public void mouseReleased(MouseEvent e)	{}
		
		public void mousePressed(MouseEvent e) {}
	
		public void setEditMode(int state){
			editMode = state;
			updateWindows();
		}
		
		public void paint(Graphics g) {
            super.paint(g);
            mag = getMagnification();
            rectOffset = canvas.getSrcRect();
            currentFrame = imp.getSlice()-1;
			int Ncenters = extractedData[currentFrame].getNregions();
			double[] cX = extractedData[currentFrame].getRegionCentroidX();
			double[] cY = extractedData[currentFrame].getRegionCentroidY();
			Polygon[] cellOutline = extractedData[currentFrame].getCellOutline();            
			
			//centroids
			g.setColor(centerColor);
        	for(int i=0; i<Ncenters; i++){
				int x0 = (int)(mag*(cX[i]-rectOffset.x)-1);
        		int y0 = (int)(mag*(cY[i]-rectOffset.y)-1);
        		g.drawLine(x0, y0-3, x0, y0+3);
        		g.drawLine(x0-3, y0, x0+3, y0);
            }
			
        	//labels
        	g.setColor(centerColor);
        	for(int i=0; i<Ncenters; i++){
        		if(labelMode==TRACKLABELMODE){
            		g.drawString(""+linkedData[currentFrame].trackLabel[i], (int)(mag*(cX[i]-rectOffset.x)-1), (int)(mag*(cY[i]-rectOffset.y)-1));
            	}else if(labelMode==CELLLABELMODE){
                	g.drawString(""+i, (int)(mag*(cX[i]-rectOffset.x)-1), (int)(mag*(cY[i]-rectOffset.y)-1));
            	}
        	}
        	
            //outlines
        	if(showOutline){
        		for(int i=0; i<Ncenters; i++){
        			if(labelMode==CELLLABELMODE){
            			g.setColor(outlineColor[i]);
            	    }else if(labelMode==TRACKLABELMODE){
            	    	g.setColor(trackLUT[linkedData[currentFrame].trackLabel[i]]);
                	}else if(labelMode==NOLABELMODE){
//            			g.setColor(Color.white);
            			g.setColor(CTExtractedData.segmentationTypeColor[extractedData[currentFrame].getSegmentationType()[i]]);
                	}
            		if(cellOutline[i] !=null){
	            		int npoints = cellOutline[i].npoints;
	            		int[] xpoints = new int[npoints];
	            		int[] ypoints = new int[npoints];
	            		for(int point=0; point<npoints; point++){
	            			xpoints[point] =  (int)(mag*(cellOutline[i].xpoints[point]-rectOffset.x));
	            			ypoints[point] =  (int)(mag*(cellOutline[i].ypoints[point]-rectOffset.y));
	    	            }
	            		g.drawPolygon(xpoints, ypoints, npoints);
	            	}
	        	}
            }
        	
        	//tracking
        	g.setColor(pairingColor);
            if(displayMode==PAIRINGMODE){
            	if((currentFrame+1) != extractedData.length){
            		double[] cX2 = extractedData[currentFrame+1].getRegionCentroidX();
    				double[] cY2 = extractedData[currentFrame+1].getRegionCentroidY();
    				
    				for(int i=0; i<Ncenters; i++){
    	            	int Ndaughters = linkedData[currentFrame].children[i].size();
    	            	for(int j=0; j<Ndaughters; j++){
    	            		int xd1 = (int)((cX[i]-rectOffset.x)*mag);
    	            		int yd1 = (int)((cY[i]-rectOffset.y)*mag);
    	            		int xd2 = (int)((cX2[linkedData[currentFrame].children[i].elementAt(j).intValue()]-rectOffset.x)*mag);
    	            		int yd2 = (int)((cY2[linkedData[currentFrame].children[i].elementAt(j).intValue()]-rectOffset.y)*mag);
    	            		drawArrow(g, xd1, yd1, xd2, yd2);
    	    			}
    				}	
    				if(editMode==ADDCONECTIONSTATE){
    					int Ncenters2 = extractedData[currentFrame+1].getNregions();
    					for(int i=0; i<Ncenters2; i++){
    		            	g.fillOval((int)(mag*(cX2[i]-rectOffset.x)-1), (int)(mag*(cY2[i]-rectOffset.y)-1), 3, 3);
    		         	}
    				}
            	}
        	}else if(displayMode==TRACKMODE){
        		int MAXDEPTH = Integer.MAX_VALUE;
        		for(int i=0; i<Ncenters; i++){
        		    boolean trackEnded = false;			
        		    int depth = 0;
        		    int ilocal = i;
        		    g.setColor(trackLUT[linkedData[currentFrame].trackLabel[i]%MAXCOLORS]);
        		    while((currentFrame-depth)>0 && depth<MAXDEPTH && !trackEnded){
        				double[] cX1 = extractedData[currentFrame-depth].getRegionCentroidX();
        				double[] cY1 = extractedData[currentFrame-depth].getRegionCentroidY();
        				
                		double[] cX2 = extractedData[currentFrame-depth-1].getRegionCentroidX();
        				double[] cY2 = extractedData[currentFrame-depth-1].getRegionCentroidY();
        				
        				
            			int x1 = (int)(mag*(cX1[ilocal]-rectOffset.x)-1);
            			int y1 = (int)(mag*(cY1[ilocal]-rectOffset.y)-1);
            			
            			int[] mother = linkedData[currentFrame-depth].mother;

            			if(mother[ilocal]!=-1){
            				int x2 = (int)(mag*(cX2[mother[ilocal]]-rectOffset.x)-1);
                			int y2 = (int)(mag*(cY2[mother[ilocal]]-rectOffset.y)-1);
                			ilocal = mother[ilocal];
                			g.drawLine(x1, y1, x2, y2);
            			}else{
            				trackEnded = true;
            			}
        				depth++;
        			}
        		}
        	}
		}
		
		private void drawArrow(Graphics g, int xd1, int yd1, int xd2, int yd2) {
			double length = Math.sqrt((xd1-xd2)*(xd1-xd2)+(yd1-yd2)*(yd1-yd2));
    		double theta = Math.atan2(yd2-yd1, xd2-xd1);
			double barb = 3.0*length/16.0;
	    	double angle = Math.PI/9.0;
	    
	    	g.drawLine(xd1, yd1, xd2, yd2);
	    	
	        int x = (int)Math.round(xd2 + barb*Math.cos(angle+theta+Math.PI));
	        int y = (int)Math.round(yd2 + barb*Math.sin(angle+theta+Math.PI));

	    	g.drawLine(x, y, xd2, yd2);

	        x = (int)Math.round(xd2 + barb*Math.cos(-angle+theta+Math.PI));
	        y = (int)Math.round(yd2 + barb*Math.sin(-angle+theta+Math.PI));

	    	g.drawLine(x, y, xd2, yd2);
	    }
	}
}
