/*====================================================================
| Version: November 11, 2010
\===================================================================*/

/*====================================================================
| Ricard Delgado-Gonzalo
| EPFL/STI/IOA/LIB/BM.4.141
| e-mail: ricard.delgado@gmail.com
| URL: http://www.ricarddelgado.com/
\===================================================================*/
package filters;
import ij.process.ImageProcessor;


public class CTMinMax {

    public final static int MIN = 0;
    public final static int MAX = 1;
    
	public CTMinMax(){
		
	}
	
	public void filter2DMINMAX(ImageProcessor ip, int type, int xRadius, int yRadius ){
        if (xRadius>0) {
        	filter1DMINMAX(ip, type, xRadius, true, yRadius);
        }
        if (yRadius>0) {
        	filter1DMINMAX(ip, type, yRadius, false, 0);
        }
    }
    
    private void filter1DMINMAX(ImageProcessor ip, int type, int radius, boolean xDirection, int extraY) {
        float sign = (type==MIN) ? -1 : 1;
        int width = ip.getWidth();
        int height = ip.getHeight();
     
        float[] pixels = (float[])ip.getPixels();
        int length = xDirection ? width : height;   //number of points per line (line can be a row or column)
        float[] cache = new float[length];          //input for filter, hopefully in CPU cache
        int pointInc = xDirection ? 1 : width;      //increment of the pixels array index to the next point in a line
        int lineInc = xDirection ? width : 1;       //increment of the pixels array index to the next line
        int lineTo = (xDirection ? height : width); //the last line+1 to process
        int writeTo = xDirection ? width : height;
        int readTo = (writeTo+radius > length) ? length : writeTo+radius;
        int pixel0 = 0; //the first pixel to write in a line
        for (int line=0; line<lineTo; line++, pixel0+=lineInc) {
            int p = line*lineInc;
            for (int i=0; i<readTo; i++ ,p+=pointInc){
                cache[i] = pixels[p]*sign;
            }
            lineMax(radius, sign, cache, pixels, 0, writeTo, pixel0, pointInc);
        }
    }

    private static void lineMax (int radius, float sign, float[] cache, float[] pixels, int writeFrom, int writeTo, int pixel0, int pointInc) {
        int length = cache.length;
        int pUp = pixel0;
        int pDn = pixel0 + (writeTo-1)*pointInc;
        float maxUp = -Float.MAX_VALUE;
        float maxDn = -Float.MAX_VALUE;
        int iInUp = radius;         //new in the range that we have to find the max of
        int iOutUp = - radius - 1;    //not in the range any more
        int iInDn = writeTo - radius -1;
        int iOutDn = writeTo + radius;

        while (pUp<=pDn) {
            boolean first = true;
            for (; pUp<=pDn; pUp+=pointInc, iInUp++, iOutUp++) {
                float oldmax = maxUp;
                if (iInUp<length && maxUp<cache[iInUp]) maxUp = cache[iInUp];
                if (first || (iOutUp>=0 && cache[iOutUp]==oldmax && (iInUp>=length || cache[iOutUp]>cache[iInUp]))) {
                if (!first) break;
                    int maxFrom = (iOutUp >= -1) ? iOutUp+1 : 0;
                    int maxTo = (iInUp < length) ? iInUp : length-1;
                    maxUp = cache[maxFrom];
                    for (int i=maxFrom+1; i<=maxTo; i++)
                        if (maxUp < cache[i]) maxUp = cache[i];
                }
            
                first = false;
                pixels[pUp] = maxUp*sign;
            }
            first = true;
            for (; pUp<=pDn; pDn-=pointInc, iInDn--, iOutDn--) {
                float oldmax = maxDn;
                if (iInDn>=0 && maxDn<cache[iInDn]) maxDn = cache[iInDn];
                if (first || (iOutDn<length && cache[iOutDn]==oldmax && (iInDn<0 || cache[iOutDn]>cache[iInDn]))) {
                    if (!first) break;
                    int maxFrom = (iOutDn <=length) ? iOutDn-1 : length-1;
                    int maxTo = (iInDn > 0) ? iInDn : 0;
                    maxDn = cache[maxFrom];
                    for (int i=maxFrom-1; i>=maxTo; i--)
                        if (maxDn < cache[i]) maxDn = cache[i];
                }
                first = false;
                pixels[pDn] = maxDn*sign;
            }
        }
    }
	
}
