package filters;
/*====================================================================
| Version: December 7, 2010
\===================================================================*/

/*====================================================================
| Ricard Delgado-Gonzalo
| EPFL/STI/IOA/LIB/BM.4.141
| e-mail: ricard.delgado@gmail.com
| URL: http://www.ricarddelgado.com/
\===================================================================*/
import ij.process.FloatProcessor;

public class CTQuadraticSteerableJFilter{
    private double sigma;
    private static final int M = 4;
    private static final double[] w4 = {0.059, -0.204, 0.024, -0.194, 0.063};

    private CTSteerableDetector detector = null;
	
    public CTQuadraticSteerableJFilter(FloatProcessor ip, double sigma){
		int nx = ip.getWidth();
		int ny = ip.getHeight();
		int nz = 1;
		this.sigma = sigma;
		
		int size = nx*ny*nz; 
		
		float[] input_f = (float[])ip.getPixels(); 
		double[] input = new double[size];
		for(int i=0; i<size; i++){
			input[i] = (double)input_f[i];
		}
		
		detector = new CTSteerableDetector(input, nx, ny, nz, sigma, M, getAlpha());
    }
	
	public void filter(){
		detector.filter();
	}

	public void showNMS(){
		detector.showNMS();
	}
	
	public void showResult(){
		detector.showResponse("");
	}
	
	public double[] getNMS(){
		return(detector.computeNMS());
	} 
	
	private double[] getAlpha() {
		double[] alpha = new double[5];
		double sigma2 = sigma*sigma;
		alpha[0] = w4[0]*sigma;
		alpha[1] = w4[1]*sigma;
		alpha[2] = w4[2]*sigma*sigma2;
		alpha[3] = w4[3]*sigma*sigma2;
		alpha[4] = w4[4]*sigma*sigma2;
		return alpha;
	}		
}

