/*====================================================================
| Version: November 11, 2010
\===================================================================*/

/*====================================================================
| Ricard Delgado-Gonzalo
| EPFL/STI/IOA/LIB/BM.4.141
| e-mail: ricard.delgado@gmail.com
| URL: http://www.ricarddelgado.com/
\===================================================================*/
package segmentation;

import ij.IJ;

import java.awt.Point;
import java.util.Vector;

public class CTContourTracing {
	private boolean[] T;
	private int nPoints;
	private int width;
	private int height;
	private Vector<Point> path;
	private static final int UP = 0;
	private static final int DOWN = 1;
	private static final int LEFT = 2;
	private static final int RIGHT = 3;
	
	public CTContourTracing(boolean[] T, int width, int height){
		this.T = T;
		this.width = width;
		this.height = height;
		path = new Vector<Point>();
		nPoints = 0; 
	}

	public void trace(){
		int sx = 0;
		int sy = 0;
		int px = 0;
		int py = 0;
		int direction = UP;
		
		boolean found = false;
		for(int y=height-1; y>=0 && !found; y--){
			for(int x=0; x<width && !found; x++){
				if(T[x+y*width]){
					found = true;
					sx = x;
					sy = y;
					px = x;
					py = y;
					path.addElement(new Point(x,y));
				}
			}
		}
		
		direction = RIGHT;
		px--;
		boolean cond;
		while(!(sx==px && sy==py)){
			
			if(px<0 || px>=width || py<0 || py>=height){
				cond=false;
			}else{
				cond = T[px+py*width]; 
			}
			if(cond){
				path.addElement(new Point(px,py));
				switch(direction){
					case UP:
						px++;
						direction = LEFT;
						break;
					case DOWN:
						px--;
						direction = RIGHT;
						break;
					case LEFT:
						py--;
						direction = DOWN;
						break;
					case RIGHT: 
						py++;
						direction = UP;
						break;
					default:
						IJ.error("Orientation unknown");
						break;
				}
			}else{
				switch(direction){
				case UP:
					px--;
					direction = RIGHT;
					break;
				case DOWN:
					px++;
					direction = LEFT;
					break;
				case LEFT:
					py++;
					direction = UP;
					break;
				case RIGHT: 
					py--;
					direction = DOWN;
					break;
				default:
					IJ.error("Orientation unknown");
					break;
				}
			}
		}
		nPoints = path.size();
	}
	
	public int[] getXCoordinates(){
		int[] xCoordinates = new int[nPoints];
		for(int i=0; i<nPoints; i++){
			xCoordinates[i] = path.elementAt(i).x;
		}
		return xCoordinates;
	}
	
	public int[] getYCoordinates(){
		int[] yCoordinates = new int[nPoints];
		for(int i=0; i<nPoints; i++){
			yCoordinates[i] = path.elementAt(i).y;
		}
		return yCoordinates;
	}	
	
	public int getNPoints(){
		return nPoints;
	}
}
