/*====================================================================
| Version: November 11, 2010
\===================================================================*/

/*====================================================================
| Ricard Delgado-Gonzalo
| EPFL/STI/IOA/LIB/BM.4.141
| e-mail: ricard.delgado@gmail.com
| URL: http://www.ricarddelgado.com/
\===================================================================*/
package segmentation;
import static java.lang.Math.sqrt;
import big.ij.snake2D.Snake2D;
import big.ij.snake2D.Snake2DNode;
import big.ij.snake2D.Snake2DScale;
import ij.IJ;
import ij.process.ImageProcessor;

import java.awt.Color;
import java.awt.geom.Point2D;
 
public class CTOvuscule implements Snake2D {
	
	private boolean alive;
	private boolean convergence;
	private boolean immortalFlag;
	private int iterations;
	private int maxIterations;
	
	private float[] data;
	private boolean canceledByUser = false;
	private double area, RamanujanPerimeter;
	private double a11, a12, a22, a33, a3;
	private double c1, c2;
	private double p1, p2, pq1, pq2, pq;
	private double q1, q2, qr1, qr2, qr;
	private double r1, r2, rp1, rp2, rp;
	private double s1, s2;
	private double u1, u2;
	private double v1, v2;
	private double y1, y2;
	private double aspectRatio;
	private int width;
	private int height;
	private Snake2DNode[] node = new Snake2DNode[3];
	private static double AREA_FACTOR = 2.0 * Math.PI / Math.sqrt(27.0);
	private static double HALF_SQRT2 = Math.sqrt(2.0) / 2.0;
	private static double REGULARIZATION_WEIGHT = 100.0;
	private static double SQRT_TINY = Math.sqrt((double)Float.intBitsToFloat((int)0x33FFFFFF));
	private static double SQRT2 = Math.sqrt(2.0);
	private static double SQRT3 = Math.sqrt(3.0);
	private static double THIRD_SQRT2 = Math.sqrt(2.0) / 3.0;
	private static double ASPECT_RATIO_TH = 0.1;
	
	public CTOvuscule (ImageProcessor ip, int maxIterations, boolean immortalFlag) {
		data = (float[]) ip.getPixels();
		this.maxIterations = maxIterations;
		this.immortalFlag = immortalFlag;
		iterations = 0;
		width = ip.getWidth();
		height = ip.getHeight();
		node[0] = new Snake2DNode(width, (int)(height/2.0));
		node[1] = new Snake2DNode((int)(width/4.0),	(int)(height*(2.0-Math.sqrt(3.0))/4.0));
		node[2] = new Snake2DNode((int)(width/4.0),	(int)(height*(2.0+Math.sqrt(3.0))/4.0));
		setNodes(node);
		convergence = true;
		alive = true;
	}
	
	public void reviveSnake(){
		alive = true;
	}
	
	public void clearConvergenceStatus(){
		convergence = true;
	}
	
	public void clearIterations(){
		iterations = 0;
	}
	
	public boolean didConverge(){
		return(convergence);
	}
	
	public boolean isAlive(){
		return(alive);
	}
	
	public double energy () {
		if(!immortalFlag){
			iterations++;
			if(iterations >= maxIterations){
				alive = false;
			}
		}
		return(contrast() + regularization());
	}

	public Point2D.Double[] getEnergyGradient() {
		return(plus(contrastGradient(), regularizationGradient()));
	}
	
	public Snake2DNode[] getMNodes(int M){
		Snake2DNode[] mNodes = new Snake2DNode[M];
		for (int t = 0; t < M; t++) {
			double theta = 2.0 * Math.PI * (double)t / (double)M;
			mNodes[t] = new Snake2DNode((int)Math.round(y1 + HALF_SQRT2 * (c1 * Math.cos(theta) + s1 * Math.sin(theta))), (int)Math.round(y2 + HALF_SQRT2 * (c2 * Math.cos(theta) + s2 * Math.sin(theta))));
		}
		return(mNodes);
	}
	
	public Snake2DNode[] getNodes () {
		return(node);
	} 
	
	public Snake2DScale[] getScales() {
		int K = (int)Math.ceil(1.0 + RamanujanPerimeter / 2.0);
		if (K > (2 * (width + height))) {
			return(null);
		}
		int[] s0x1 = new int[K];
		int[] s0x2 = new int[K];
		int[] s1x1 = new int[K];
		int[] s1x2 = new int[K];
		for (int k = 0; (k < K); k++) {
			double theta = 2.0 * Math.PI * (double)k / (double)K;
			s0x1[k] = (int)Math.round(y1 + c1 * Math.cos(theta) + s1 * Math.sin(theta));
			s0x2[k] = (int)Math.round(y2 + c2 * Math.cos(theta) + s2 * Math.sin(theta));
			s1x1[k] = (int)Math.round(y1 + HALF_SQRT2 * (c1 * Math.cos(theta) + s1 * Math.sin(theta))); 
			s1x2[k] = (int)Math.round(y2 + HALF_SQRT2 * (c2 * Math.cos(theta) + s2 * Math.sin(theta)));
		}
		Snake2DScale[] skin = new Snake2DScale[2];
		skin[0] = new Snake2DScale(s0x1, s0x2, K, null, null, true, false);
		skin[1] = new Snake2DScale(s1x1, s1x2, K, Color.RED, null, true, false);
		return(skin);
	}
	
	public void setNodes (Snake2DNode[] node) {
		p1 = node[0].x;
		p2 = node[0].y;
		q1 = node[1].x;
		q2 = node[1].y;
		r1 = node[2].x;
		r2 = node[2].y;
		this.node[0].x = p1;
		this.node[0].y = p2;
		this.node[1].x = q1;
		this.node[1].y = q2;
		this.node[2].x = r1;
		this.node[2].y = r2;
		pq1 = p1 - q1;
		pq2 = p2 - q2;
		qr1 = q1 - r1;
		qr2 = q2 - r2;
		rp1 = r1 - p1;
		rp2 = r2 - p2;
		pq = p1 * q2 - p2 * q1;
		qr = q1 * r2 - q2 * r1;
		rp = r1 * p2 - r2 * p1;
		y1 = (p1 + q1 + r1) / 3.0;
		y2 = (p2 + q2 + r2) / 3.0;
		a11 = 3.0 * (p2 * pq2 + q2 * qr2 + r2 * rp2);
		a12 = 3.0 * (p1 * (q2 - 2.0 * p2) + p2 * q1 + q1 * (r2 - 2.0 * q2) + q2 * r1 + r1 * (p2 - 2.0 * r2) + r2 * p1);
		a22 = 3.0 * (p1 * pq1 + q1 * qr1 + r1 * rp1);
		a33 = pq + qr + rp;
		a3 = Math.abs(a33);
		c1 = (pq1 - rp1) / 3.0;
		c2 = (pq2 - rp2) / 3.0;
		s1 = qr1 / SQRT3;
		s2 = qr2 / SQRT3;
		double halfWidth = Math.sqrt(c1 * c1 + s1 * s1);
		double halfHeight = Math.sqrt(c2 * c2 + s2 * s2);
		u1 = y1 - halfWidth;
		u2 = y2 - halfHeight;
		v1 = y1 + halfWidth;
		v2 = y2 + halfHeight;
		double a = (a11 + a22) / 3.0;
		double b = Math.sqrt(Math.abs(a * a - 3.0 * a33 * a33));
		double semiMinor = THIRD_SQRT2 * Math.sqrt(Math.abs(a - b));
		double semiMajor = THIRD_SQRT2 * Math.sqrt(Math.abs(a + b));
		aspectRatio = semiMinor/semiMajor;
		area = AREA_FACTOR * a3;
		double excentricity = (semiMajor - semiMinor) / (semiMajor + semiMinor);
		excentricity *= excentricity;
		RamanujanPerimeter = Math.PI * (semiMajor + semiMinor) * (1.0 + 3.0 * excentricity / (10.0 + Math.sqrt(Math.abs(4.0 - 3.0 * excentricity))));
	} 
	
	public double snakeExcentricity(){
		double a = (a11 + a22) / 3.0;
		double b = Math.sqrt(Math.abs(a * a - 3.0 * a33 * a33));
		double semiMinor = THIRD_SQRT2 * Math.sqrt(Math.abs(a - b));
		double semiMajor = THIRD_SQRT2 * Math.sqrt(Math.abs(a + b));
		double excentricity = Math.sqrt(1- (semiMinor/semiMajor)*(semiMinor/semiMajor));
		return excentricity;
	}
	
	
	public void updateStatus (boolean canceledByUser, boolean snakeDied, boolean optimalSnakeFound, Double energy) {
		this.canceledByUser = canceledByUser;
		//IJ.write("END OF THE OVUSCULE STAGE");
		//if(canceledByUser) {
		//	IJ.write("Optimization process stopped by the user.");
		//	return;
		//}
		//if(snakeDied) 			IJ.write("Snake reached the maximum number of iterations.");
		//if(optimalSnakeFound) 	IJ.write("Optimal snake found.");
		//if(energy!=null) 		IJ.write("Energy "+energy.doubleValue());
	} 
	
	private double contrast () {
		if (area < 1.0) {
			convergence = false;
			alive = false;
			return(Double.MAX_VALUE);
		}
		if (aspectRatio < ASPECT_RATIO_TH){
			return(Double.MAX_VALUE);
		}
		double c = 0.0;
		int xmin = Math.max((int)Math.floor(u1), 0);
		int xmax = Math.min((int)Math.ceil(v1), width - 1);
		int ymin = Math.max((int)Math.floor(u2), 0);
		int ymax = Math.min((int)Math.ceil(v2), height - 1);
		if ((xmax <= xmin) || (ymax <= ymin)) {
			convergence = false;
			alive = false;
			return(Double.MAX_VALUE);
		}
		boolean region1Exists = false;
		for (int y = ymin; (y <= ymax); y++) {
			double dy = y2 - (double)y;
			double dy2 = dy * dy;
			for (int x = xmin; (x <= xmax); x++) {
				double dx = y1 - (double)x;
				double dx2 = dx * dx;
				double d = Math.sqrt(dx2 + dy2);
				double z = a11 * dx2 + a12 * dx * dy + a22 * dy2;
				if (z < SQRT_TINY) {
					region1Exists = true;
					c += data[x+width*y];
					continue;
				}
				z = a3 / sqrt(z);
				double d0 		= (1.0 - z) * d;
				double d0prime 	= (1.0 - z / SQRT2) * d;
				if (d0prime < -HALF_SQRT2) {
					region1Exists = true;
					c += data[x+width*y];
				}else if(-HALF_SQRT2 <= d0prime && d0prime < HALF_SQRT2){
					c -= SQRT2 * d0prime * data[x+width*y];
				}else if(HALF_SQRT2 <= d0prime && d0 < -1.0){
					c -= data[x+width*y];
				}else if(-1.0 <= d0 && d0 < 1.0){
					c -= (1.0 - d0) * data[x+width*y] / 2.0;
				}else if(-1 <= d0 && d0prime < HALF_SQRT2){
					return(Double.MAX_VALUE);
				}
			}
		}
		if(!region1Exists){
			return(Double.MAX_VALUE);		
		} 
		return(c / area);
	}
	
	private Point2D.Double[] contrastGradient () {
		Point2D.Double[] gradient = new Point2D.Double[3];
		gradient[0] = new Point2D.Double(0.0, 0.0);
		gradient[1] = new Point2D.Double(0.0, 0.0);
		gradient[2] = new Point2D.Double(0.0, 0.0);
		if (area < 1.0) {
			convergence = false;
			alive = false;
			return(gradient);
		}
		if (aspectRatio < ASPECT_RATIO_TH){
			return(gradient);
		}
		int xmin = Math.max((int)Math.floor(u1), 0);
		int xmax = Math.min((int)Math.ceil(v1), width - 1);
		int ymin = Math.max((int)Math.floor(u2), 0);
		int ymax = Math.min((int)Math.ceil(v2), height - 1);
		if ((xmax <= xmin) || (ymax <= ymin)) {
			convergence = false;
			alive = false;
			return(gradient);
		}
		double[] z12 = {pq1 - qr1, pq2 - qr2, qr1 - rp1, qr2 - rp2, rp1 - pq1, rp2 - pq2};
		boolean region1Exists = false;
		for (int y = ymin; (y <= ymax); y++) {
			double dy = y2 - (double)y;
			double dy2 = dy * dy;
			for (int x = xmin; (x <= xmax); x++) {
				double dx = y1 - (double)x;
				double dx2 = dx * dx;
				double d = sqrt(dx2 + dy2);
				if (d < SQRT_TINY) {
					continue;
				}
				double f = -data[x+width*y];
				double z = a11 * dx2 + a12 * dx * dy + a22 * dy2;
				if (z < SQRT_TINY) {
					region1Exists = true;
					gradient[0].x += f * qr2;
					gradient[0].y -= f * qr1;
					gradient[1].x += f * rp2;
					gradient[1].y -= f * rp1;
					gradient[2].x += f * pq2;
					gradient[2].y -= f * pq1;
					continue;
				}
				z = a3 / Math.sqrt(z);
				double d0 		= (1.0 - z) * d;
				double d0prime 	= (1.0 - z / SQRT2) * d;
				if (d0prime < -HALF_SQRT2) {
					region1Exists = true;
					gradient[0].x += f * qr2;
					gradient[0].y -= f * qr1;
					gradient[1].x += f * rp2;
					gradient[1].y -= f * rp1;
					gradient[2].x += f * pq2;
					gradient[2].y -= f * pq1;
				}else if (-HALF_SQRT2 <= d0prime && d0prime < HALF_SQRT2) {
					double g = SQRT2 * d;
					double g0 = z * z * z * d / (6.0 * a33);
					double g1 = a33 * (SQRT2 - z) / (3.0 * g0 * d);
					double gx = (2.0 * a11 + g1) * dx + a12 * dy;
					double gy = (2.0 * a22 + g1) * dy + a12 * dx;
					double h0 = 9.0 * (z12[5] * dx - z12[4] * dy);
					double h1 = 9.0 * (z12[1] * dx - z12[0] * dy);
					double h2 = 9.0 * (z12[3] * dx - z12[2] * dy);
					gradient[0].x -= f * (g * qr2 - g0 * (dy * h0 + gx));
					gradient[0].y += f * (g * qr1 - g0 * (dx * h0 - gy));
					gradient[1].x -= f * (g * rp2 - g0 * (dy * h1 + gx));
					gradient[1].y += f * (g * rp1 - g0 * (dx * h1 - gy));
					gradient[2].x -= f * (g * pq2 - g0 * (dy * h2 + gx));
					gradient[2].y += f * (g * pq1 - g0 * (dx * h2 - gy));
				}else if (HALF_SQRT2 <= d0prime && d0 < -1.0) {
					gradient[0].x -= f * qr2;
					gradient[0].y += f * qr1;
					gradient[1].x -= f * rp2;
					gradient[1].y += f * rp1;
					gradient[2].x -= f * pq2;
					gradient[2].y += f * pq1;
				}else if (-1.0 <= d0 && d0 < 1.0) {
					double g = (d - 1.0) / 2.0;
					double g0 = z * z * z * d / (12.0 * a33);
					double g1 = a33 * (1.0 - z) / (6.0 * g0 * d);
					double gx = (2.0 * a11 + g1) * dx + a12 * dy;
					double gy = (2.0 * a22 + g1) * dy + a12 * dx;
					double h0 = 9.0 * (z12[5] * dx - z12[4] * dy);
					double h1 = 9.0 * (z12[1] * dx - z12[0] * dy);
					double h2 = 9.0 * (z12[3] * dx - z12[2] * dy);
					gradient[0].x += f * (g * qr2 - g0 * (dy * h0 + gx));
					gradient[0].y -= f * (g * qr1 - g0 * (dx * h0 - gy));
					gradient[1].x += f * (g * rp2 - g0 * (dy * h1 + gx));
					gradient[1].y -= f * (g * rp1 - g0 * (dx * h1 - gy));
					gradient[2].x += f * (g * pq2 - g0 * (dy * h2 + gx));
					gradient[2].y -= f * (g * pq1 - g0 * (dx * h2 - gy));
				}else if(-1 <= d0 && d0prime < HALF_SQRT2){
					gradient[0] = new Point2D.Double(0.0, 0.0);
					gradient[1] = new Point2D.Double(0.0, 0.0);
					gradient[2] = new Point2D.Double(0.0, 0.0);
					return(gradient);
				}
			}
		}
		if(!region1Exists){
			gradient[0] = new Point2D.Double(0.0, 0.0);
			gradient[1] = new Point2D.Double(0.0, 0.0);
			gradient[2] = new Point2D.Double(0.0, 0.0);
			return(gradient);
		} 
		double A = a33 * area;
		gradient[0].x /= A;
		gradient[0].y /= A;
		gradient[1].x /= A;
		gradient[1].y /= A;
		gradient[2].x /= A;
		gradient[2].y /= A;
		return(gradient);
	}
	
	private Point2D.Double[] plus (Point2D.Double[] gc, Point2D.Double[] gr	) {
		if ((gc == null) || (gr == null)) {
			IJ.error("Ovuscule Class Internal Error");
			return(null);
		}
		int K = gc.length;
		if (K != gr.length) {
			IJ.error("Ovuscule Class Internal Error");
			return(null);
		}
		Point2D.Double[] g = new Point2D.Double[K];
		for (int k = 0; (k < K); k++) {
			g[k] = new Point2D.Double(gc[k].x + gr[k].x, gc[k].y + gr[k].y);
		}
		return(g);
	}
	
	private double regularization () {
		double regularization = Math.min(Math.min(pq2 * pq2, qr2 * qr2), rp2 * rp2);
		return(REGULARIZATION_WEIGHT * regularization / area);
	} 
	
	private Point2D.Double[] regularizationGradient () {
		Point2D.Double[] gradient = new Point2D.Double[3];
		gradient[0] = new Point2D.Double(0.0, 0.0);
		gradient[1] = new Point2D.Double(0.0, 0.0);
		gradient[2] = new Point2D.Double(0.0, 0.0);
		double ppqq2 = pq2 * pq2;
		double qqrr2 = qr2 * qr2;
		double rrpp2 = rp2 * rp2;
		double A = 0.0;
		if ((ppqq2 <= qqrr2) && (ppqq2 <= rrpp2)) {
			gradient[0].x = -pq2 * qr2;
			gradient[0].y = 2.0 * a33 + pq2 * qr1;
			gradient[1].x = -rp2 * pq2;
			gradient[1].y = -2.0 * a33 + rp1 * pq2;
			gradient[2].x = -ppqq2;
			gradient[2].y = pq1 * pq2;
			A = REGULARIZATION_WEIGHT * pq2 / (a33 * area);
		}else if ((qqrr2 <= rrpp2) && (qqrr2 <= ppqq2)) {
			gradient[0].x = -qqrr2;
			gradient[0].y = qr1 * qr2;
			gradient[1].x = -qr2 * rp2;
			gradient[1].y = 2.0 * a33 + qr2 * rp1;
			gradient[2].x = -pq2 * qr2;
			gradient[2].y = -2.0 * a33 + pq1 * qr2;
			A = REGULARIZATION_WEIGHT * qr2 / (a33 * area);
		}else if ((rrpp2 <= ppqq2) && (rrpp2 <= qqrr2)) {
			gradient[0].x = -qr2 * rp2;
			gradient[0].y = -2.0 * a33 + qr1 * rp2;
			gradient[1].x = -rrpp2;
			gradient[1].y = rp1 * rp2;
			gradient[2].x = -rp2 * pq2;
			gradient[2].y = 2.0 * a33 + rp2 * pq1;
			A = REGULARIZATION_WEIGHT * rp2 / (a33 * area);
		}
		gradient[0].x *= A;
		gradient[0].y *= A;
		gradient[1].x *= A;
		gradient[1].y *= A;
		gradient[2].x *= A;
		gradient[2].y *= A;
		return(gradient);
	}

	public boolean isCanceledByUser() {
		return (canceledByUser);
	}
}