/*====================================================================
| Version: November 11, 2010
\===================================================================*/

/*====================================================================
| Ricard Delgado-Gonzalo
| EPFL/STI/IOA/LIB/BM.4.141
| e-mail: ricard.delgado@gmail.com
| URL: http://www.ricarddelgado.com/
\===================================================================*/
package segmentation;
import ij.IJ;
import ij.ImagePlus;
import ij.ImageStack;
import ij.process.ByteProcessor;
import ij.process.ColorProcessor;
import ij.process.FloatProcessor;

import java.util.Vector;

import javax.swing.JTextArea;


public class CTWatershed extends Object {

	static private int MAXBASINS = 1000000;
	static private int CONNECTIVITY = 8;
	static private float MASK = 2.0f;
	static private float WSHED = 0.0f;
	static private float INQUEUE = 3.0f;
	static public float FIRSTLABEL = 10.0f;
	static private float INIT = 1.0f;
	
	Vector<Integer> fifo = new Vector<Integer>();
	private int nx, ny;
	private boolean stop = false;
	
	private ImageStack stackAnimation;
	
	private byte LUTColor[][];
	
	private int minLevel;
	private int maxLevel;
	private boolean animation = false;
	private float[] outPixels;
	private int[] inPixels;  
	private FloatProcessor output = null;
	private JTextArea taskLog;
	
	CTWatershed() {
		this.LUTColor = CTWatershedDisplay.createLUTColor(MAXBASINS);
	}
	
	public void setLog(JTextArea taskLog){
		this.taskLog = taskLog;
	}
	
	public void enableAnimation() {
		animation = true;
	}
	
	private void writeLog(String str){
		if(taskLog!=null){
			taskLog.append(str+"\n");
			taskLog.setCaretPosition(taskLog.getDocument().getLength());
		}else{
			IJ.log(str);
		}
	}
	
	public FloatProcessor doWatershed(ByteProcessor image) {
		writeLog("Starting the watershed flooding...");
		nx = image.getWidth();
		ny = image.getHeight();
		inPixels = new int[nx*ny];
		byte[] inPixels_tmp = (byte[])image.getPixels();  
		
		for(int i=0; i<nx*ny; i++){
			inPixels[i] = ((byte)inPixels_tmp[i] & 0xFF);
		}
		writeLog("Computing image range...");

		int index;
		minLevel = Integer.MAX_VALUE;
		maxLevel = 0;
		for(int i=1; i<nx-1; i++){ 
			for(int j=1; j<ny-1; j++) {
		 		index = i + j*nx;
		 		if (inPixels[index] < minLevel){ 
			 		minLevel = inPixels[index];
			 	}
		 		if (inPixels[index] > maxLevel){ 
					maxLevel = inPixels[index];
		 		}
			}
		}

		writeLog("Image range ["+minLevel+","+maxLevel+"]");
		writeLog("Computing histogram...");
		
		int h[];								// histogram
		int pos[];								// position of the first pixel of a certain gray level in the sorted array
		double pixelValue[] = new double[nx*ny];// gray value of the sorted pixel
		int pixelX[] = new int[nx*ny];			// original x position of the sorted pixel
		int pixelY[] = new int[nx*ny];			// original y position of the sorted pixel
		int pixelPos[][] = new int[nx][ny];		// sorted position of the pixel (for reverse access)
		
		stop = false;
		h = hist();
		
		pos = initPos(h);
		sort(pos, pixelValue, pixelX, pixelY, pixelPos);
		
		writeLog("Performing the flooding...");
		
		output = flooding(h, pos, pixelValue, pixelX, pixelY, pixelPos);
		return output;
	}

	public void stop() {
		stop = true;
	}
		
	private int[] hist() {
		int v;
		int[] h = new int[256];
		for (int i=0; i<256;i++){
			h[i]=0;
		}
		for (int x=1; x<nx-1; x++){
			for (int y=1; y<ny-1; y++) {
				v = (int)inPixels[x+nx*y];
				h[v]++;
			}
		}
		return h;
	}
	
	private int[] initPos(int[] h) {
		int pos[] = new int[256];
		pos[0]=0;
		pos[1]=h[0];
		for(int i=2; i<256; i++){
			pos[i] = h[i-1] + pos[i-1];
		}
		return pos;
	}
	
	private void sort(int pos[], double pixelValue[], int pixelX[], int pixelY[], int pixelPos[][]) {
		int v;
		int len;
		
		len = pos.length;
		int position[] = new int[len];
		
		for (int i = 0; i < len; i++){
			position[i] = pos[i];
		}
		
		for (int x=1; x<nx-1; x++){
			for (int y=1; y<ny-1; y++){
				v = inPixels[x+y*nx];
				pixelValue[position[v]] = v;
				pixelX[position[v]] = x;
				pixelY[position[v]] = y;
				pixelPos[x][y] = position[v];
				position[v]++;
			}
		}
		
	}
	
	
	private FloatProcessor flooding(int[] h, int[] pos,	double[] pixelValue, int[] pixelX, int[] pixelY, int[][] pixelPos) {
		FloatProcessor output = new FloatProcessor(nx, ny);;
		outPixels = (float[])output.getPixels();
		
		int size = nx*ny;
		int limit = (nx-2)*(ny-2);
		
		float neigh[] = new float[CONNECTIVITY];
		int gray = (int)pixelValue[0];
		boolean flag = false;
		float current_label = FIRSTLABEL;
		
		int dx[] = new int[CONNECTIVITY];
		int dy[] = new int[CONNECTIVITY];
		dx[0] = dx[3] = dx[5] = -1;
		dx[2] = dx[4] = dx[7] = 1;
		dx[1] = dx[6] = 0;
		dy[0] = dy[1] = dy[2] = -1;
		dy[5] = dy[6] = dy[7] = 1;
		dy[3] = dy[4] = 0;
		
		for (int k = 0; k < size; k++)
			outPixels[k] = INIT;
		
		int xp, yp;
		if (animation){
			createStackAnimation();
		}
		
		float central;
		int pgray, p;
		int loopEnd = maxLevel-minLevel+1;
		for (int l = 1; l <= loopEnd && stop == false; l++) {
			pgray = pos[gray];
			for (int j = 0; j < h[gray]; j++) {
				xp = pixelX[pgray+j];
				yp = pixelY[pgray+j];
				outPixels[xp+nx*yp] = MASK;
				getNeighborhood(xp, yp, neigh);
				for (int n=0; n<CONNECTIVITY; n++) {
					if (neigh[n] >= FIRSTLABEL || neigh[n] == WSHED) {
						outPixels[xp+nx*yp] = INQUEUE;
						fifo.addElement(new Integer(pgray+j));
						break;
					}
				}
			}
			//int fs = fifo.size();
			while (!fifo.isEmpty()) {
				p = ((Integer)fifo.elementAt(0)).intValue();
				fifo.removeElementAt(0);
				xp = pixelX[p];
				yp = pixelY[p];
				central = getNeighborhood(xp, yp, neigh);
				for (int n=0; n<CONNECTIVITY; n++) {
					if (neigh[n] >= FIRSTLABEL) {
						if ((central == INQUEUE) || ((central == WSHED) && (flag == true))) {
							outPixels[xp+nx*yp] = neigh[n];
							central = neigh[n];
						}else if ((central >= FIRSTLABEL) && (central != neigh[n])) {
							outPixels[xp+nx*yp] = WSHED;
							central = WSHED;
							flag = false;
						}
					}else if(neigh[n] == WSHED){
						if (central == INQUEUE) {
							outPixels[xp+nx*yp] = WSHED;
							central = WSHED;
							flag = true;
						}
					}else if (neigh[n] == MASK) {
						outPixels[(xp+dx[n])+nx*(yp+dy[n])] = INQUEUE;
						neigh[n] = INQUEUE;
						fifo.addElement(new Integer(pixelPos[xp+dx[n]][yp+dy[n]]));
					}
				}
			}
				
			pgray = pos[gray];
			for (int j = 0; (j < h[gray]) && (stop == false); j++) {
				xp = pixelX[pgray+j];
				yp = pixelY[pgray+j];
				if (outPixels[xp+nx*yp] == MASK) {
					current_label++;
					stop = (current_label > MAXBASINS);
					fifo.addElement(new Integer(pgray+j));
					outPixels[xp+nx*yp] =  current_label;
					while (!fifo.isEmpty() && (stop == false)){
						p = ((Integer)fifo.elementAt(0)).intValue();
						fifo.removeElementAt(0);
						getNeighborhood(pixelX[p], pixelY[p], neigh);
						for (int n=0; n<CONNECTIVITY; n++){
							if (neigh[n] == MASK) {
								xp = pixelX[p]+dx[n];
								yp = pixelY[p]+dy[n];
								fifo.addElement(new Integer(pixelPos[xp][yp]));
								outPixels[xp+nx*yp] = current_label;
								neigh[n] = current_label;
							}
						}
					}
				}
			}

			if (pos[gray]+h[gray] < limit){
				gray = (int)pixelValue[pos[gray]+h[gray]];
			}
			
			if (animation){
				addImageToStackAnimation(output);
			}
		}
		
		if (animation)
			displayStackAnimation();
		
		FloatProcessor outws = (FloatProcessor) output.duplicate();
		float[] outwsPixels = (float[])outws.getPixels();
		
		int index;
		for(int i=1; i<nx-1; i++){ 
			for(int j=1; j<ny-1; j++) {
				index = i+j*nx;
				if (outPixels[index] < FIRSTLABEL) {
					if (outPixels[index+1] > FIRSTLABEL)
						outwsPixels[index] = WSHED;
					else if (outPixels[index-1] > FIRSTLABEL)
						outwsPixels[index] = WSHED;
					else if (outPixels[index+nx]  > FIRSTLABEL)
						outwsPixels[index] = WSHED;
					else if (outPixels[index-nx]  > FIRSTLABEL)
						outwsPixels[index] = WSHED;
				}
			}
		}
						
		int i, j;
		j = (ny-1)*nx;
		for(i=0; i<nx; i++) {
			outwsPixels[i] = WSHED;
			outwsPixels[i+j] = WSHED;
		}
		i = nx-1;
		for(j=0; j<size; j+=nx) {
			outwsPixels[j] = WSHED;
			outwsPixels[i+j] = WSHED;
		}
		
		
		
		return outws;
	}
	
	/*
	 * Get a 4-connected or 8-connected neighborhood
	 *
	 * Arragement of the pixel in 4-connect
	 *	In neigh[0] there is the pixel(x, y-1)
	 *	In neigh[1] there is the pixel(x-1, y)
	 *	In neigh[2] there is the pixel(x+1, y)
	 *	In neigh[3] there is the pixel(x, y+1)
	 *
	 * Arragement of the pixel in 8-connect
	 *	In neigh[0] there is the pixel(x-1, y-1)
	 *	In neigh[1] there is the pixel(x-1, y)
	 *	In neigh[2] there is the pixel(x-1, y+1)
	 *	In neigh[3] there is the pixel(x, y-1)
	 *	In neigh[4] there is the pixel(x, y+1)
	 *	In neigh[5] there is the pixel(x+1, y-1)
	 *	In neigh[6] there is the pixel(x+1, y)
	 *	In neigh[7] there is the pixel(x+1, y+1)
	 *
	 *	return the central pixel
	 */
	private float getNeighborhood(int x, int y, float[] neigh) {
		int index;
		float central;
		index = x-1 + (y-1)*nx;
		neigh[0] = outPixels[index];	index++;
		neigh[1] = outPixels[index];	index++;
		neigh[2] = outPixels[index];	index += (nx-2);
		neigh[3] = outPixels[index];	index++;
		central  = outPixels[index];	index++;
		neigh[4] = outPixels[index];	index += (nx-2);
		neigh[5] = outPixels[index];	index++;
		neigh[6] = outPixels[index];	index++;
		neigh[7] = outPixels[index];
		return central;
	}
	
	/**
	 */
	private void createStackAnimation() {
		stackAnimation = new ImageStack(nx, ny);
	}
	
	/**
	 */
	private void addImageToStackAnimation(FloatProcessor image) {
		int size = nx*ny;
		float pixels[] = (float[])image.getPixels();
		byte r[] = new byte[size];
		byte g[] = new byte[size];
		byte b[] = new byte[size];
		int v;
		for (int k = 0; k < size; k++) {
			v = (int)pixels[k];
			if (v < 0)
				v = 0;
			else if (v > MAXBASINS)
				v = MAXBASINS;
			r[k] = LUTColor[0][v];
			g[k] = LUTColor[1][v];
			b[k] = LUTColor[2][v];
		}
		ColorProcessor cp = new ColorProcessor(nx, ny);
		cp.setRGB(r, g, b);
		stackAnimation.addSlice("", cp);
	}
	
	/**
	 */
	private void displayStackAnimation() {
		ImagePlus animation = new ImagePlus("Animation", stackAnimation);
		animation.show();
	}
	
}

