/*====================================================================
| Version: November 11, 2010
\===================================================================*/

/*====================================================================
| Ricard Delgado-Gonzalo
| EPFL/STI/IOA/LIB/BM.4.141
| e-mail: ricard.delgado@gmail.com
| URL: http://www.ricarddelgado.com/
\===================================================================*/
package segmentation;
import ij.ImagePlus;
import ij.process.ByteProcessor;
import ij.process.ColorProcessor;
import ij.process.FloatProcessor;
import ij.process.ImageProcessor;

public class CTWatershedDisplay {
	
	static public byte[][] createLUTColor(int MAXBASINS) {
		byte LUTColor[][] = new byte[3][MAXBASINS];
		for (int c=0; c<3; c++) {
			for (int n=0; n < 10; n++)
				LUTColor[c][n] = 0;
			for (int n=10; n < MAXBASINS; n++)
				LUTColor[c][n] = (byte)(10+235*Math.random());
		}
		return LUTColor;
	}
	
	static public void showBasins(FloatProcessor output, byte[][] LUTColor) {
		int nx = output.getWidth();
		int ny = output.getHeight();
		int size = nx*ny;
		FloatProcessor r = new FloatProcessor(nx, ny);
		FloatProcessor g = new FloatProcessor(nx, ny);
		FloatProcessor b = new FloatProcessor(nx, ny);
		int v;
		float[] pixr = (float[])r.getPixels();
		float[] pixg = (float[])g.getPixels();
		float[] pixb = (float[])b.getPixels();
		float[] pixout = (float[])output.getPixels();
		for (int index=0; index<size; index++) {
			v = (int)pixout[index];
			pixr[index] = LUTColor[0][v];
			pixg[index] = LUTColor[1][v];
			pixb[index] = LUTColor[2][v];
		}
		ColorProcessor cp = new ColorProcessor(nx, ny);
		cp.setPixels(0,(FloatProcessor)r); 
		cp.setPixels(1,(FloatProcessor)g); 
		cp.setPixels(2,(FloatProcessor)b); 
		ImagePlus imp_tmp = new ImagePlus("Colorized basins", cp);
		imp_tmp.resetDisplayRange();
		imp_tmp.show();	
	}

	static public void showBinary(FloatProcessor output) {
		int nx = output.getWidth();
		int ny = output.getHeight();
		ByteProcessor out = new ByteProcessor(nx, ny);

		for (int x=0; x<nx; x++){
			for (int y=0; y<ny; y++) {
				if ( output.getPixel(x, y) < 2)
					out.putPixel(x, y, 255);
				if ( output.getPixel(x-1, y) < 2)
					out.putPixel(x, y-1, 255);
				if ( output.getPixel(x, y) < 2)
					out.putPixel(x+1, y, 255);
				if ( output.getPixel(x, y+1) < 2)
					out.putPixel(x, y, 255);
			}
		}
		ImagePlus imp_tmp = new ImagePlus("Binary watershed lines", out);
		imp_tmp.show();
	}
	
	static public void showDams(FloatProcessor output) {
		int nx = output.getWidth();
		int ny = output.getHeight();
		ByteProcessor out = new ByteProcessor(nx, ny);
		
		for (int x=0; x<nx; x++){
			for (int y=0; y<ny; y++){
				if (output.getPixel(x, y) == 0){
					out.putPixel(x, y, 0);
				}else{
					out.putPixel(x, y, 255);
				}
			}
		}
		ImagePlus imp_tmp = new ImagePlus("Watershed lines", out);
		imp_tmp.show();
	}
	
	static public void showLabelizedBasins(FloatProcessor output){
		ImageProcessor output_tmp = output.duplicate();
		output_tmp.resetMinAndMax();
		ImagePlus imageplus = new ImagePlus("Labelized basins",output_tmp);
		imageplus.show();
	}

	static public void showGravityCenters(FloatProcessor output, double[] barycenterX, double[] barycenterY) {
		int nx = output.getWidth();
		int ny = output.getHeight();
		int nb = barycenterX.length;
		FloatProcessor r = new FloatProcessor(nx, ny);
		float[] pixr = (float[])r.getPixels();
		int index;
		for(int i=0; i<nb; i++){
			index = (int)Math.round(barycenterX[i]) + (int)Math.round(barycenterY[i])*nx;
			pixr[index] = (float)255;
		}
		ImagePlus imp_tmp = new ImagePlus("Gravity centers", r);
		imp_tmp.show();
		
	}
}