/*====================================================================
| Version: November 11, 2010
\===================================================================*/

/*====================================================================
| Ricard Delgado-Gonzalo
| EPFL/STI/IOA/LIB/BM.4.141
| e-mail: ricard.delgado@gmail.com
| URL: http://www.ricarddelgado.com/
\===================================================================*/
package structures;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.geom.Point2D;
import java.util.Vector;

import javax.swing.JFrame;

import display.CTLineDeviationChart;
import display.CTLinesDeviationChart;
import display.CTStackedBarChart;
import ij.IJ;

public class CTStatistics {
	
	private Vector<JFrame> frames = null;
	
	public CTStatistics(){
		frames = new Vector<JFrame>();
	}
	
	private Vector<Point2D.Double> computeQoSMean(CTExtractedData[] extractedData, int[] countPerFrame, int type){
		Vector<Point2D.Double> meanPerformance = new Vector<Point2D.Double>();
		int nFrames = extractedData.length;		
		
		for(int i=0; i<nFrames; i++){
			if(countPerFrame[i]!=0){
				int nRegions = extractedData[i].getNregions();
				int[] segmentationType = extractedData[i].getSegmentationType();
				double[] segmentationQuality = extractedData[i].getSegmentationQuality();
				double meanPerformanceElement = 0;
				for(int j=0; j<nRegions; j++){
					if(segmentationType[j]==type){
						meanPerformanceElement += segmentationQuality[j];
					}
				}
				meanPerformanceElement /= (double)countPerFrame[i];
				meanPerformance.addElement(new Point2D.Double((double)i+1,meanPerformanceElement));
			}
		}
		return meanPerformance;
	}
	
	private Vector<Point2D.Double> computeCircMean(CTExtractedData[] extractedData, int[] countPerFrame, int type){
		Vector<Point2D.Double> meanPerformance = new Vector<Point2D.Double>();
		int nFrames = extractedData.length;		
		
		for(int i=0; i<nFrames; i++){
			if(countPerFrame[i]!=0){
				int nRegions = extractedData[i].getNregions();
				int[] segmentationType = extractedData[i].getSegmentationType();
				double[] segmentationQuality = extractedData[i].getCellCircularity();
				double meanPerformanceElement = 0;
				for(int j=0; j<nRegions; j++){
					if(segmentationType[j]==type){
						meanPerformanceElement += segmentationQuality[j];
					}
				}
				meanPerformanceElement /= (double)countPerFrame[i];
				meanPerformance.addElement(new Point2D.Double((double)i+1,meanPerformanceElement));
			}
		}
		return meanPerformance;
	}
	
	private Vector<Point2D.Double> computeQoSStd(CTExtractedData[] extractedData, int[] countPerFrame, int type, Vector<Point2D.Double> meanPerformance){
		Vector<Point2D.Double> stdPerformance = new Vector<Point2D.Double>();
		int length = meanPerformance.size();		
		for(int i=0; i<length; i++){
			int frame = (int)Math.round(meanPerformance.elementAt(i).x-1);
			int nRegions = extractedData[frame].getNregions();
			int[] segmentationType = extractedData[frame].getSegmentationType();
			double[] segmentationQuality = extractedData[frame].getSegmentationQuality();
			double stdPerformanceElement = 0;
			for(int j=0; j<nRegions; j++){
				if(segmentationType[j]==type){
					stdPerformanceElement += (segmentationQuality[j]-meanPerformance.elementAt(i).y)*(segmentationQuality[j]-meanPerformance.elementAt(i).y);
				}
			}
			stdPerformanceElement = Math.sqrt(stdPerformanceElement/(double)(countPerFrame[i]-1));
			stdPerformance.addElement(new Point2D.Double((double)frame+1,stdPerformanceElement));
			
		}
		return stdPerformance;
	}

	private Vector<Point2D.Double> computeCircStd(CTExtractedData[] extractedData, int[] countPerFrame, int type, Vector<Point2D.Double> meanPerformance){
		Vector<Point2D.Double> stdPerformance = new Vector<Point2D.Double>();
		int length = meanPerformance.size();		
		for(int i=0; i<length; i++){
			int frame = (int)Math.round(meanPerformance.elementAt(i).x-1);
			int nRegions = extractedData[frame].getNregions();
			int[] segmentationType = extractedData[frame].getSegmentationType();
			double[] segmentationQuality = extractedData[frame].getCellCircularity();
			double stdPerformanceElement = 0;
			for(int j=0; j<nRegions; j++){
				if(segmentationType[j]==type){
					stdPerformanceElement += (segmentationQuality[j]-meanPerformance.elementAt(i).y)*(segmentationQuality[j]-meanPerformance.elementAt(i).y);
				}
			}
			stdPerformanceElement = Math.sqrt(stdPerformanceElement/(double)(countPerFrame[i]-1));
			stdPerformance.addElement(new Point2D.Double((double)frame+1,stdPerformanceElement));
			
		}
		return stdPerformance;
	}
	
	@SuppressWarnings("unchecked")
	public void segmentationPerformance(CTExtractedData[] extractedData){
		int[] esnakes 	= countSegmentationMode(extractedData, CTExtractedData.ESNAKE);
		int[] ovuscules = countSegmentationMode(extractedData, CTExtractedData.OVUSCULE);
		int[] regions 	= countSegmentationMode(extractedData, CTExtractedData.REGION);
		
		int[][] segmentationTypes = new int[3][];
		segmentationTypes[0] = esnakes;
		segmentationTypes[1] = ovuscules;
		segmentationTypes[2] = regions;
		
		String[] legend = new String[3];
		legend[0] = new String("E-snakes");
		legend[1] = new String("Ovuscules");
		legend[2] = new String("Region");
			
		CTStackedBarChart chartCounter = new CTStackedBarChart("Segmentation Assessment", segmentationTypes, legend);
		chartCounter.setPreferredSize(new Dimension(900,400));
		chartCounter.pack();
		chartCounter.setVisible(true);
		
		Vector<Point2D.Double> esnakesMeanPerformance 	= computeQoSMean(extractedData, esnakes, CTExtractedData.ESNAKE);
		Vector<Point2D.Double> ovusculesMeanPerformance = computeQoSMean(extractedData, ovuscules, CTExtractedData.OVUSCULE);
		Vector<Point2D.Double> regionsMeanPerformance 	= computeQoSMean(extractedData, regions, CTExtractedData.REGION);
		
		Vector<Point2D.Double> esnakesStdPerformance 	= computeQoSStd(extractedData, regions, CTExtractedData.ESNAKE, esnakesMeanPerformance);
		Vector<Point2D.Double> ovusculesStdPerformance 	= computeQoSStd(extractedData, regions, CTExtractedData.OVUSCULE, ovusculesMeanPerformance);
		Vector<Point2D.Double> regionsStdPerformance 	= computeQoSStd(extractedData, regions, CTExtractedData.REGION, regionsMeanPerformance);
		
		CTLineDeviationChart chartEsnakes = new CTLineDeviationChart("E-Snakes Mean Energy", "Energy", Color.red, esnakesMeanPerformance, esnakesStdPerformance);
		chartEsnakes.setPreferredSize(new Dimension(1000,400));
		chartEsnakes.pack();
		chartEsnakes.setVisible(true);
		
		CTLineDeviationChart chartOvuscules = new CTLineDeviationChart("Ovuscules Mean Energy", "Energy", Color.yellow, ovusculesMeanPerformance, ovusculesStdPerformance);
		chartOvuscules.setPreferredSize(new Dimension(1000,400));
		chartOvuscules.pack();
		chartOvuscules.setVisible(true);

		CTLineDeviationChart chartRegion = new CTLineDeviationChart("Cell Area vs. Convex Hull Cell Area", "Ratio", Color.blue, regionsMeanPerformance, regionsStdPerformance);
		chartRegion.setPreferredSize(new Dimension(1000,400));
		chartRegion.pack();
		chartRegion.setVisible(true);
		
		
		Vector<Point2D.Double> esnakesCircularity 	= computeCircMean(extractedData, esnakes, CTExtractedData.ESNAKE);
		Vector<Point2D.Double> ovusculesCircularity = computeCircMean(extractedData, ovuscules, CTExtractedData.OVUSCULE);
		Vector<Point2D.Double> regionsCircularity 	= computeCircMean(extractedData, regions, CTExtractedData.REGION);
		
		Vector<Point2D.Double> esnakesStdCircularity = computeCircStd(extractedData, regions, CTExtractedData.ESNAKE, esnakesCircularity);
		Vector<Point2D.Double> ovusculesStdCircularity = computeCircStd(extractedData, regions, CTExtractedData.OVUSCULE, ovusculesCircularity);
		Vector<Point2D.Double> regionsStdCircularity = computeCircStd(extractedData, regions, CTExtractedData.REGION, regionsCircularity);
		
		Vector<Point2D.Double>[] meanCircularities = new Vector[3];
		meanCircularities[0] = esnakesCircularity;
		meanCircularities[1] = ovusculesCircularity;
		meanCircularities[2] = regionsCircularity;

		Vector<Point2D.Double>[] stdCircularities = new Vector[3];
		stdCircularities[0] = esnakesStdCircularity;
		stdCircularities[1] = ovusculesStdCircularity;
		stdCircularities[2] = regionsStdCircularity;
		
		CTLinesDeviationChart chartCircularities = new CTLinesDeviationChart("Region Circularity", "Circularity", meanCircularities, stdCircularities, legend);
		chartCircularities.setPreferredSize(new Dimension(1000,400));
		chartCircularities.pack();
		chartCircularities.setVisible(true);
				
		frames.addElement(chartEsnakes);
		frames.addElement(chartOvuscules);
		frames.addElement(chartRegion);
		frames.addElement(chartCounter);
		frames.addElement(chartCircularities);
	}
	
	public void disposeAll(){
		int nFrames = frames.size();
		for(int i=0; i<nFrames; i++){
			frames.elementAt(i).dispose();
		}
		frames.clear();
	}
	
	private int[] countSegmentationMode(CTExtractedData[] extractedData, int type){
		int nFrames = extractedData.length;
		if(nFrames<=0){
			IJ.error("Incorrect number of frames");
			return null;
		}
		if(type != CTExtractedData.ESNAKE && type != CTExtractedData.OVUSCULE && type != CTExtractedData.REGION ){
			IJ.error("Segmentation method unknown");
			return null;
		}
		
		int[] array = new int[nFrames];
		
		for(int i=0; i<nFrames; i++){
			int nRegions = extractedData[i].getNregions();
			int counter = 0;
			int[] segmentationType = extractedData[i].getSegmentationType();
			for(int j=0; j<nRegions; j++){
				if(segmentationType[j] == type){
					counter++;
				}
			}
			array[i] = counter;
		}
		return array;
	}
	
	
}
