/*====================================================================
| Version: November 11, 2010
\===================================================================*/

/*====================================================================
| Ricard Delgado-Gonzalo
| EPFL/STI/IOA/LIB/BM.4.141
| e-mail: ricard.delgado@gmail.com
| URL: http://www.ricarddelgado.com/
\===================================================================*/
package tracking;

import java.util.Arrays;
import java.util.Vector;

import structure.CTEdge;
import structure.CTMatching;
import tracking.CTBellmanFord;


public class CTMaximumWeightMatching {
	private int Unodes;
	private int Nnodes;
	private Vector<CTEdge> maximumWeightMatching;
	
	@SuppressWarnings("unchecked")
	public CTMaximumWeightMatching(Vector<CTEdge> G, int Unodes, int Wnodes) {
		this.Unodes = Unodes;
		Nnodes = Unodes+Wnodes;
		
		Vector<CTEdge> GM, M, path;
		Vector<CTMatching> Mset;
		
		Mset	= new Vector<CTMatching>();
		GM		= new Vector<CTEdge>();
		M		= new Vector<CTEdge>();
		path	= new Vector<CTEdge>();
		
		boolean[] matched = new boolean[Nnodes];
		Arrays.fill(matched, false);
		String str = new String("");
		do{
			for(int i=0; i<G.size(); i++){
				boolean found = false;
				int j=0;
				for(j=0; j<M.size() && !found; j++){
					if(G.elementAt(i).equals(M.elementAt(j))){
						found = true;
						matched[G.elementAt(i).source] = true;
						matched[G.elementAt(i).destination] = true;
						GM.add(new CTEdge(G.elementAt(i).destination, G.elementAt(i).source, G.elementAt(i).weight));
					}
				}
				if(!found) GM.add(new CTEdge(G.elementAt(i).source, G.elementAt(i).destination, -G.elementAt(i).weight));
			}
			
			path = findBestAugmentingPath(GM, matched);
			GM.removeAllElements();
			
			if(path.size()!=0){
				for(int i=0; i<path.size(); i++){
					boolean found = false;
					int j=0;
					for(j=0; j<M.size() && !found; j++){
						if(path.elementAt(i).equals(M.elementAt(j))){
							found = true;
							M.removeElementAt(j);
						}
					}
					if(!found){ 
						M.add(new CTEdge(path.elementAt(i).source, path.elementAt(i).destination, path.elementAt(i).weight));
						matched[path.elementAt(i).source] = true;
						matched[path.elementAt(i).destination] = true;
					}
				}
			}
			str += "\t"+ matchingWeight(M);
			Mset.addElement(new CTMatching((Vector<CTEdge>) M.clone(), matchingWeight(M)));
		}while(path.size()!=0);
		//IJ.write(str);
		
		//identify best matching
		int bestIndex = 0;
		double bestWeight = 0.0;
		
		for(int i=0; i< Mset.size(); i++){
			double currentWeight = Mset.elementAt(i).weight;
			if(currentWeight>bestWeight){
				bestWeight = currentWeight;
				bestIndex = i;
			}
		}
		
		maximumWeightMatching = Mset.elementAt(bestIndex).cTEdges;
		
	}

	public Vector<CTEdge> getMaximumWeightMatching(){
		return maximumWeightMatching;
		
	}
	
	private double matchingWeight(Vector<CTEdge> M){
		double w = 0;
		int size = M.size();
		for(int i=0; i<size; i++){
			w += Math.abs(M.elementAt(i).weight);
		}
		return w;
	}
	
	
	@SuppressWarnings("unchecked")
	private Vector<CTEdge> findBestAugmentingPath(Vector<CTEdge> GM, boolean[] matched){
		Vector<CTEdge> G = (Vector<CTEdge>) GM.clone(); 
		
		for(int i=0; i<Nnodes; i++){
			if(!matched[i]){
				if(i<Unodes){
					G.add(new CTEdge(Nnodes, i, 0.0));
				}else{
					G.add(new CTEdge(i,Nnodes+1,  0.0));
				}
			}
		}
		
		CTBellmanFord trackerBellmanFord = new CTBellmanFord(G, Nnodes+2, Nnodes);
		trackerBellmanFord.computeSpanningTree();

		int[] predecessor = trackerBellmanFord.getPredecessor();
		double[] weights = trackerBellmanFord.getTotalWeight();
		
		Vector<CTEdge> path = new Vector<CTEdge>();
	
		int n1 = predecessor[Nnodes+1]; 
		
		
		
		
		while(n1!=Nnodes && n1!=-1){
			int n2 =  predecessor[n1];
			if(predecessor[n2]!=-1){
				if(n1<n2){
					path.add(new CTEdge(n1,n2,Math.abs(weights[n1]-weights[n2])));
				}else{
					path.add(new CTEdge(n2,n1,Math.abs(weights[n1]-weights[n2])));
				}
			}	
			n1=n2;
		}
		
		return path;
		
	}
	
}

