/*====================================================================
| Version: November 11, 2010
\===================================================================*/

/*====================================================================
| Ricard Delgado-Gonzalo
| EPFL/STI/IOA/LIB/BM.4.141
| e-mail: ricard.delgado@gmail.com
| URL: http://www.ricarddelgado.com/
\===================================================================*/
package tracking;

import ij.IJ;

import java.util.Vector;

import javax.swing.JProgressBar;
import javax.swing.JTextArea;

import structure.CTEdge;
import structures.CTConnectivity;
import structures.CTExtractedData;
import structures.CTRunParameters;

public class CTTracking {
	private int labelTmp;
	private int Nstack;
	private CTRunParameters runParameters = null;
	private CTExtractedData[] extractedData = null;
	private CTConnectivity[] linkedData = null;
	private int[][] offset8connect = null; 
	private JProgressBar progressBar = null;
	private JTextArea taskLog = null;
	
	public CTTracking(CTRunParameters runParameters, CTExtractedData[] extractedData, CTConnectivity[] linkedData){
		this.runParameters = runParameters;
		this.extractedData = extractedData;
		this.linkedData = linkedData;
		Nstack = extractedData.length;
		
		offset8connect = new int[2][9];
		offset8connect[0][0] = 0;
		offset8connect[1][0] = 0;
		
		offset8connect[0][1] = 0;
		offset8connect[1][1] = 1;
		
		offset8connect[0][2] = 1;
		offset8connect[1][2] = 0;

		offset8connect[0][3] = 1;
		offset8connect[1][3] = 1;

		offset8connect[0][4] = 0;
		offset8connect[1][4] = -1;

		offset8connect[0][5] = -1;
		offset8connect[1][5] = 0;

		offset8connect[0][6] = -1;
		offset8connect[1][6] = -1;
		
		offset8connect[0][7] = -1;
		offset8connect[1][7] = 1;
		
		offset8connect[0][8] = 1;
		offset8connect[1][8] = -1;
	}
	
	public void setProgressBar(JProgressBar progressBar){
		this.progressBar = progressBar;
	}
	
	public void setLog(JTextArea taskLog){
		this.taskLog = taskLog;
	}
	
	private void writeLog(String str){
		if(taskLog!=null){
			taskLog.append(str+"\n");
			taskLog.setCaretPosition(taskLog.getDocument().getLength());
		}else{
			IJ.log(str);
		}
	}
	
	public void doTracking(){
		int nFiles1 = Nstack-1;
		for(int i=0; i<nFiles1; i++) {
			graphBasedTracking(i, i+1);
			if(progressBar!=null) progressBar.setValue(i+1);
		}
		labelTracks();
	}
	
	private void updateWeightsOnGraph(Vector<CTEdge> matching, Vector<CTEdge> graph, Vector<CTEdge> similarityGraph, int[] frameIndexToGraphIndex, int[] graphIndexToFrameIndex, CTExtractedData trackerFrameExtractedData1,	CTExtractedData trackerFrameExtractedData2) {
		
		writeLog("Updating matching flow model compatibility...");
		double[] centroidX1 = trackerFrameExtractedData1.getRegionCentroidX();
		double[] centroidY1 = trackerFrameExtractedData1.getRegionCentroidY();
		double[] centroidX2 = trackerFrameExtractedData2.getRegionCentroidX();
		double[] centroidY2 = trackerFrameExtractedData2.getRegionCentroidY();
		
		int Nregions1 = trackerFrameExtractedData1.getNregions();
		boolean[][] neighborhoodMatrix1 = trackerFrameExtractedData1.getNeighborhoodMatrix();
		
		int NGraphIndex = graphIndexToFrameIndex.length;
		int Nedges = graph.size();
		int MatchingSize = matching.size();
		
		double[] regularization = new double[Nedges];
		//add regularization
		writeLog("Computing N-regularization...");
		for(int i=0; i<Nedges; i++){
			int index1Graph = graph.elementAt(i).source;
			int frame1Index = graphIndexToFrameIndex[index1Graph];
			int frame2Index = graph.elementAt(i).destination-NGraphIndex;
			
			double vx =  centroidX2[frame2Index] - centroidX1[frame1Index];
			double vy =  centroidY2[frame2Index] - centroidY1[frame1Index];
			regularization[i] = 0.0;
			int Nvalid = 0;
			
			//loop in all neighborhood matrix to find the neighbors of frame1Index
			for(int k=0; k<Nregions1; k++){
				if(neighborhoodMatrix1[frame1Index][k]){
					int kGraph = frameIndexToGraphIndex[k]; //neighbor found
					
					if(kGraph!=-1){//valid element?
						boolean lFound = false;
						for(int l=0; l<MatchingSize && !lFound; l++){ //find in the matching to where it goes
							int sourceIndexGraph = matching.elementAt(l).source;
							if(kGraph==sourceIndexGraph){
								lFound=true;
								Nvalid++;

								int destinationFrameIndex = matching.elementAt(l).destination-NGraphIndex;
								double dx = (centroidX2[destinationFrameIndex]-centroidX1[graphIndexToFrameIndex[sourceIndexGraph]]) - vx;
								double dy = (centroidY2[destinationFrameIndex]-centroidY1[graphIndexToFrameIndex[sourceIndexGraph]]) - vy;
								dx = dx*dx;
								dy = dy*dy;
								regularization[i] += dx+dy;
							}
						}
					}
				}
			}
			
			if(Nvalid!= 0){ 
				regularization[i] /= (double)Nvalid;
			}
			
		}
		
		double maxDisplacement = Math.max(runParameters.lateralDisplacement, Math.max(runParameters.upDisplacement, runParameters.downDisplacement));
		
		for(int i=0; i<Nedges; i++){
			graph.elementAt(i).weight = similarityGraph.elementAt(i).weight -runParameters.lambda*Math.sqrt(regularization[i])/maxDisplacement;
		}
		
		writeLog("Matching flow model compatibility updated");
	}
	
	private void loadSimilarityWeightsOnGraph(Vector<CTEdge> graph, int[] frameIndexToGraphIndex, int[] graphIndexToFrameIndex, CTExtractedData trackerFrameExtractedData1, CTExtractedData trackerFrameExtractedData2) {
		writeLog("Computing similarity weights...");

		//extract important information
		int[] widthRegion1 = trackerFrameExtractedData1.getRegionWidth();
		int[] heightRegion1 = trackerFrameExtractedData1.getRegionHeight();

		int[] widthRegion2 = trackerFrameExtractedData2.getRegionWidth();
		int[] heightRegion2 = trackerFrameExtractedData2.getRegionHeight();
			
		double[][] targetImageValues1 = trackerFrameExtractedData1.getRegionImageValuesCh1();
		double[][] targetImageValues2 = trackerFrameExtractedData2.getRegionImageValuesCh1();
		
		int NGraphIndex = graphIndexToFrameIndex.length;
		int Nedges = graph.size();
		double[] correlations = new double[Nedges];
		for(int i=0; i<Nedges; i++){
			int frame1Index = graphIndexToFrameIndex[graph.elementAt(i).source];
			int frame2Index = graph.elementAt(i).destination-NGraphIndex;
			correlations[i] = findBestCorrelation(targetImageValues1[frame1Index], widthRegion1[frame1Index], heightRegion1[frame1Index], targetImageValues2[frame2Index], widthRegion2[frame2Index], heightRegion2[frame2Index]);
		}
		
		for(int i=0; i<Nedges; i++){
			graph.elementAt(i).weight = -Math.sqrt(correlations[i])/(255.0);
		}
		writeLog("Similarity weights computed");
	}
	
	@SuppressWarnings("unchecked")
	private void graphBasedTracking(int frame1, int frame2){
		//initialize parameters
		int frame1Size = extractedData[frame1].getNregions();
		int frame2Size = extractedData[frame2].getNregions();
		
		double[] x1 = extractedData[frame1].getRegionCentroidX();
		double[] y1 = extractedData[frame1].getRegionCentroidY();
		double[] x2 = extractedData[frame2].getRegionCentroidX();
		double[] y2 = extractedData[frame2].getRegionCentroidY();
		
		//building index tables
		int[] frameIndexToGraphIndex = new int[frame1Size];
		int NGraphIndex = 0;
		for(int i=0; i<frame1Size; i++){
			if(!(x1[i]<runParameters.lateralDisplacement || x1[i]>(runParameters.roiBox.width-runParameters.lateralDisplacement) || y1[i]<runParameters.upDisplacement || y1[i]>(runParameters.roiBox.height-runParameters.downDisplacement))) {
				frameIndexToGraphIndex[i] = NGraphIndex;
				NGraphIndex++;
			}else{
				frameIndexToGraphIndex[i] = -1;
			}
		}
		
		int[] graphIndexToFrameIndex = new int[NGraphIndex];
		int ii=0;
		for(int i=0; i<frame1Size; i++){
			if(!(x1[i]<runParameters.lateralDisplacement || x1[i]>(runParameters.roiBox.width-runParameters.lateralDisplacement) || y1[i]<runParameters.upDisplacement || y1[i]>(runParameters.roiBox.height-runParameters.downDisplacement))) {
				graphIndexToFrameIndex[ii] = i;
				ii++;
			}
		}
		
		//BUILD THE STRUCTURE OF THE BIPARTITE GRAPH TO MATCH
		// number of nodes of layer 1 = NGraphIndex
		// number of nodes of layer 2 = frame2Size
		
		Vector<CTEdge> similarityGraph = new Vector<CTEdge>(); //bipartite graph
		for(int i=0; i<NGraphIndex; i++){
			int i2 = graphIndexToFrameIndex[i];
			for(int j=0; j<frame2Size; j++){
				if(Math.abs(x2[j]-x1[i2])<runParameters.lateralDisplacement){
					double DeltaY = y2[j]-y1[i2]; 
					if((DeltaY>0 && DeltaY<runParameters.downDisplacement) || (DeltaY<0 && -DeltaY<runParameters.upDisplacement)){
						similarityGraph.addElement(new CTEdge(i,NGraphIndex+j,0));
					}
				}
			}
		}
		
		//compute similarity weights
		loadSimilarityWeightsOnGraph(similarityGraph, frameIndexToGraphIndex, graphIndexToFrameIndex, extractedData[frame1], extractedData[frame2]);
		
		//solve the maximum weighted matching problem
		writeLog("Solving the weighted graph matching problem with similarity measure only...");
		CTMaximumWeightMatching trackerMaximumWeightMatching = new CTMaximumWeightMatching(similarityGraph, NGraphIndex, frame2Size);
		Vector<CTEdge> matching = trackerMaximumWeightMatching.getMaximumWeightMatching();
        writeLog("Weighted graph matching solved");
		
		int matchingIter = 0;
		Vector<CTEdge> graph = (Vector<CTEdge>) similarityGraph.clone();
		while(matchingIter<runParameters.regularizationIter){
			updateWeightsOnGraph(matching, graph, similarityGraph, frameIndexToGraphIndex, graphIndexToFrameIndex, extractedData[frame1], extractedData[frame2]);
			writeLog("Solving the weighted graph matching problem with the mixed model...");
			trackerMaximumWeightMatching = new CTMaximumWeightMatching(graph, NGraphIndex, frame2Size);
			matching = trackerMaximumWeightMatching.getMaximumWeightMatching();
	        writeLog("Weighted graph matching solved");
			matchingIter++;
	    }
		
		//store the matching information
        writeLog("Storing matching...");
        int MatchingSize = matching.size();
        for(int i=0; i<MatchingSize; i++){
			int index1 = matching.elementAt(i).source;
			int index2 = matching.elementAt(i).destination-NGraphIndex;
			
			int realIndex1 = graphIndexToFrameIndex[index1];
			//backward connection
			linkedData[frame2].mother[index2] = realIndex1;
			//forward connection
			linkedData[frame1].children[realIndex1].addElement(new Integer(index2));
		}
	}
		
	private double findBestCorrelation(double[] data1, int width1, int height1, double[] data2, int width2, int height2){
		int Nlevel = (int) Math.round(Math.log(Math.max(width1, height1))/Math.log(3)); 
		
		double nodeX1 = (double)width1/2.0;
		double nodeY1 = (double)height1/2.0;
		double stepX = (double)width1/3.0;
		double stepY = (double)height1/3.0;
		double x2 = (double)width2/2.0;
		double y2 = (double)height2/2.0;

		
		double bestMSE = Double.MAX_VALUE;
		for(int level =0; level < Nlevel; level++){
			int bestOffestCounter = 0;
			double x1, y1, mse;
			int counter, u, v, xmin, ymin, xmax, ymax;
			
			for(int offsetCounter = 0; offsetCounter<9; offsetCounter++){
				x1 = nodeX1+offset8connect[0][offsetCounter]*stepX;
				y1 = nodeY1+offset8connect[1][offsetCounter]*stepY;

				u = (int)Math.round(x1-x2);
				v = (int)Math.round(y1-y2);
				
				xmin = (int)Math.ceil(Math.max(0, u));
				ymin = (int)Math.ceil(Math.max(0, v));
				
				xmax = (int)Math.floor(Math.min(width1-1, width2-1+u));
				ymax = (int)Math.floor(Math.min(height1-1,height2-1+v));
				
				mse = 0;
				counter = 0;
				for(int x=xmin; x<=xmax; x++){
					for(int y=ymin; y<=ymax; y++){
						mse += (data1[x+width1*y] - data2[x-u+width2*(y-v)])*(data1[x+width1*y] - data2[x-u+width2*(y-v)]);
						counter++;
					}			
				}			
				if(counter!=0) mse/=counter;
		
				if(mse < bestMSE){
					bestMSE = mse;
					bestOffestCounter = offsetCounter;
				}
			}
			nodeX1 += offset8connect[0][bestOffestCounter]*stepX;
			nodeY1 += offset8connect[1][bestOffestCounter]*stepY;
			stepX /= 3.0;
			stepY /= 3.0;
		}
		return bestMSE;
	}
	
	
	
	
	
	private void labelTracks(){
		int Nregions;
		//initialize
		for(int i=0; i<Nstack; i++){
			Nregions = linkedData[i].Nregions;
			linkedData[i].trackLabel = new int[Nregions];
			for(int j=0; j<Nregions; j++){
				linkedData[i].trackLabel[j] = -1;
			}
		}
		
		//label in recursive way
		labelTmp = 0;
		for(int frame=0; frame<Nstack; frame++){
			Nregions = linkedData[frame].Nregions;
			for(int cell=0; cell<Nregions; cell++){
				if(linkedData[frame].trackLabel[cell]==-1){ 
					labelTrack(frame, cell, labelTmp);
					labelTmp++;
				}
			}
		}
	}	
	
	private void labelTrack(int frame, int cell, int label) {
		linkedData[frame].trackLabel[cell] = label;
		int Ndaughters = linkedData[frame].children[cell].size();
		if(Ndaughters != 0){
			for(int i=0; i<Ndaughters; i++){
				labelTrack(frame+1, linkedData[frame].children[cell].elementAt(i), label);
			}
		}		
	}	
	
}
